#
# Copyright 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
from __future__ import absolute_import

import json
import ssl
import urllib3

from six import PY3

from virtwho.virt.kubevirt import config


class KubeClient:

    def __init__(self, path):
        cfg = config.Configuration()
        cl = config._get_kube_config_loader_for_yaml_file(path)
        cl.load_and_set(cfg)

        cert_reqs = ssl.CERT_REQUIRED
        ca_certs = cfg.ssl_ca_cert
        cert_file = cfg.cert_file
        key_file = cfg.key_file

        self._pool_manager = urllib3.PoolManager(
            num_pools=4,
            maxsize=4,
            cert_reqs=cert_reqs,
            ca_certs=ca_certs,
            cert_file=cert_file,
            key_file=key_file
        )

        self.host = cfg.host
        self.token = cfg.token
        self._version = self._kubevirt_version()

    def get_nodes(self):
        return self._request('/api/v1/nodes')

    def get_vms(self):
        return self._request('/apis/kubevirt.io/' + self._version + '/virtualmachineinstances')

    def _kubevirt_version(self):
        versions = self._request('/apis/kubevirt.io')
        return versions['preferredVersion']['version']

    def _request(self, path):
        header_params = {}

        header_params['Accept'] = 'application/json'
        header_params['Content-Type'] = 'application/json'
        header_params['Authorization'] = self.token

        url = self.host + path

        try:
            r = self._pool_manager.request("GET", url,
                                          fields=None,
                                          preload_content=True,
                                          headers=header_params)
        except urllib3.exceptions.SSLError as e:
            msg = "{0}\n{1}".format(type(e).__name__, str(e))
            raise ApiException(status=0, reason=msg)

        # In the python 3, the response.data is bytes.
        # we need to decode it to string.
        if PY3:
            data = r.data.decode('utf8')
        else:
            data = r.data

        if not 200 <= r.status <= 299:
            raise ApiException(http_resp=r)

        # fetch data from response object
        try:
            data = json.loads(data)
        except ValueError:
            data = r.data

        return data


class ApiException(Exception):

    def __init__(self, status=None, reason=None, http_resp=None):
        if http_resp:
            self.status = http_resp.status
            self.reason = http_resp.reason
            self.body = http_resp.data
            self.headers = http_resp.getheaders()
        else:
            self.status = status
            self.reason = reason
            self.body = None
            self.headers = None

    def __str__(self):
        """
        Custom error messages for exception
        """
        error_message = "({0})\n"\
                        "Reason: {1}\n".format(self.status, self.reason)
        if self.headers:
            error_message += "HTTP response headers: {0}\n".format(self.headers)

        if self.body:
            error_message += "HTTP response body: {0}\n".format(self.body)

        return error_message



from __future__ import print_function

import os
from multiprocessing import Process, Value
from six.moves.SimpleHTTPServer import SimpleHTTPRequestHandler
import random
from six.moves import socketserver
import requests
import json


class FakeServer(socketserver.TCPServer):
    allow_reuse_address = True


class FakeHandler(SimpleHTTPRequestHandler):
    def write_file(self, directory, filename):
        """
        Send file with given `filename` to the client. File must be in
        `directory` in the data/ subdirectory of the directory
        where the current __file__ is.
        """
        base = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(base, 'data', directory, filename)

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.send_header("Content-length", os.stat(path).st_size)
        self.end_headers()
        with open(path, "rb") as f:
            self.wfile.write(f.read())

    def write_json(self, content, status_code=requests.codes.ok, headers=None):
        """
        Send back as json the content parameter (Most useful to input a dict).
        Sets the following headers if not provided (other wise they are added to the input headers)
          "Content-type" : "application/json"
          "Content-length" : len(json.dumps(content).encode('utf-8'))
        :param content: The object to convert to json and send as the body of the message
        :type content: dict

        :param status_code: The HTTP Status code to set on the response
        :type status_code: int

        :param headers: A dict of header_name : header_value.
        :type headers: dict
        :return: Nothing
        :rtype: None
        """

        if headers is None:
            headers = {}
        if "Content-type" not in headers:
            headers["Content-type"] = "application/json"

        content = json.dumps(content).encode('utf-8')
        if "Content-length" not in headers:
            headers["Content-length"] = len(content)

        self.send_response(status_code)
        for header, value in headers.items():
            self.send_header(header, value)
        self.end_headers()
        self.wfile.write(content)


class FakeVirt(Process):
    def __init__(self, handler_class, host='localhost', port=None):
        super(FakeVirt, self).__init__()
        self._port = port
        self.host = host
        self.handler_class = handler_class
        self.server = FakeServer((self.host, self.port), handler_class)
        self.daemon = True
        self._data_version = Value('d', 0)

    @property
    def port(self):
        if self._port is None:
            self._port = random.randint(8000, 9000)
        return self._port

    def clear_port(self):
        print("Clear port: ", self._port)
        self._port = None

    @property
    def username(self):
        return 'A!bc3#\'"'

    @property
    def password(self):
        return 'A!bc3#\'"'

    @property
    def data_version(self):
        return self._data_version.value

    @data_version.setter
    def data_version(self, version):
        self._data_version.value = version

    def run(self):
        for i in range(100):
            try:
                print("Starting {cls} on port {port}".format(cls=self.__class__.__name__, port=self.port))
                self.server.serve_forever()
                break
            except AssertionError:
                self.clear_port()
        else:
            raise AssertionError("No free port found, starting aborted")



import json
import time
from . import ahv_constants
from requests import Session
from requests.exceptions import ConnectionError, ReadTimeout
from virtwho import virt

class AhvInterface(object):
  """ AHV REST Api interface class"""
  NO_RETRY_HTTP_CODES = [400, 404, 500, 502, 503]
  event_types = ['node', 'vm']

  def __init__(self, logger, url, username, password, port, **kwargs):
    """
    Args:
      logger (Log): Logger.
      url (str): Rest server url.
      username (str): Username.
      password (str): Password for rest client.
      port (int): Port number for ssp.
      kwargs(dict): Accepts following arguments:
        timeout(optional, int): Max seconds to wait before HTTP connection
        times-out. Default 30 seconds.
        retries (optional, int): Maximum number of retires. Default: 5.
        retry_interval (optional, int): Time to sleep between retry intervals.
        internal_debug (optional, bool): Detail log of the rest calls.
         Default: 5 seconds.
    """
    self._session = Session()
    self._timeout = kwargs.get('timeout', 30)
    self._retries = kwargs.get('retries', 5)
    self._retry_interval = kwargs.get('retry_interval', 30)
    self._logger = logger
    self._url = url
    self._user = username
    self._password = password
    self._port = port
    self._internal_debug = kwargs.get('internal_debug', False)
    self._create_session(self._user, self._password)

  def _create_session(self, user=None, password=None):
    """
    Creates rest session.
    Args:
      user (str): Username.
      password (str): Password for rest session.
    Returns:
      None.
    """
    if user is None:
      user = self._user
    if password is None:
      password = self._password
    self._session.auth = (user, password)

  def _make_url(self, uri, *args):
    """
    Creates base url.
    uri would always begin with a slash
    Args:
      uri (str): Uri.
      args (list): Args.
    Returns:
      url (str): Url with uri.
    """
    if not uri.startswith("/"):
      uri = "/%s" % uri
    url = "%s%s" % (self._url, uri)
    for arg in args:
      url += "/%s" % str(arg)
    return url

  def _format_response(self, data):
    """
    Format the data based on the response's version.
    Args:
      data (dict): Data dictionary.
    Returns:
      formatted_data (dict): Formatted dictionary.
    """
    if 'entities' in data:
      return self._process_entities_list(data['entities'])
    else:
      return self._process_dict_response(data)

  def _process_dict_response(self, data):
    """
    Format the data when we only have a dictionary.
    Args:
      data (dict): Data dictionary.
    Returns:
      formatted_data (dict): Formatted data.
    """
    formatted_data = data
    if 'status' in data and 'metadata' in data:
      formatted_data = dict(data['status'], **data['metadata'])

    if 'resources' in formatted_data:
      if 'power_state' in formatted_data['resources']:
        formatted_data['power_state'] = \
          formatted_data['resources']['power_state']
      if 'num_cpu_sockets' in formatted_data['resources']:
        formatted_data['num_cpu_sockets'] = \
          formatted_data['resources']['num_cpu_sockets']

    return formatted_data

  def _process_entities_list(self, data):
    """
    Format data for the list of entities.
    Args:
      data (list): List of entities dictionary.
    Returns:
      formatted_data (dict): Formatted data after processing list fo entities.
    """
    formatted_data = data
    initial = True
    for entity in data:
      if 'status' in entity and 'metadata' in entity:
        if initial:
          formatted_data = []
          initial = False
        formatted_data.append(dict(entity['status'], **entity['metadata']))

    for ent_obj in formatted_data:
      if 'resources' in ent_obj:
        if 'nodes' in ent_obj['resources']:
          nodes = ent_obj['resources']['nodes']
          if 'hypervisor_server_list' in nodes:
            ent_obj['hypervisor_types'] = []
            for server in nodes['hypervisor_server_list']:
              ent_obj['hypervisor_types'].append(server['type'])

      if 'kind' in ent_obj:
        if ent_obj['kind'] == 'cluster':
          if 'uuid' in ent_obj:
            ent_obj['cluster_uuid'] = ent_obj['uuid']

    return formatted_data

  def login(self, version):
    """
    Login to the rest server and ensure connection succeeds.
    Args:
      version (Str): Interface version.
    Returns:
      None.
    """
    (url, cmd_method) = self.get_diff_ver_url_and_method(
      cmd_key='list_clusters', intf_version=version)
    self.make_rest_call(method=cmd_method, uri=url)
    self._logger.info("Successfully logged into the AHV REST server")

  def get_hypervisor_type(self, version, host_entity=None, vm_entity=None):
    """
    Get the hypervisor type of the guest vm.
    Args:
      version (Str): API version.
      host_entity (Dict): Host info dict.
      vm_entity (Dict): Vm info dict.
    Returns:
      hypervisor_type (str): Vm hypervisor type.
    """
    hypervisor_type = None
    if version == 'v2.0': 
      if host_entity:
        hypervisor_type = host_entity['hypervisor_type']
      else:
        self._logger.warning("Cannot retrieve the host type. Version:%s"
                             % version)
    else:
      if vm_entity:
        if 'resources' in vm_entity:
          if 'hypervisor_type' in vm_entity['resources']:
            hypervisor_type = vm_entity['resources']['hypervisor_type']
          else:
            self._logger.debug("Hypervisor type of the %s is not available"
                               % vm_entity['uuid'])
      else:
        self._logger.warning("No vm entity is provided for version %s. "
                             "Therefore it's unable to retrieve host type"
                             % version)
    return hypervisor_type

  def get_common_ver_url_and_method(self, cmd_key):
    """
    Gets the correct cmd name based on its corresponding version.
    Args:
      cmd_key (str): Key name to search for in the command dict.
    Returns:
      (str, str) : Tuple of (command, rest_type).
    """
    return (ahv_constants.CMN_RST_CMD[cmd_key]['url'],
            ahv_constants.CMN_RST_CMD[cmd_key]['method'])

  def get_diff_ver_url_and_method(self, cmd_key, intf_version):
    """
    Gets the correct cmd name based on its corresponding version
    Args:
      cmd_key (str): Key name to search for in the command dict.
      intf_version (str): Interface version.
    Returns:
      (str, str) : Tuple of (command, rest_type).
    """
    return (ahv_constants.REST_CMD[intf_version][cmd_key]['url'],
            ahv_constants.REST_CMD[intf_version][cmd_key]['method'])

  def get(self, uri, *args, **kwargs):
    """
    Args are appended to the url as components.
    /arg1/arg2/arg3
    Send a get request with kwargs to the server.
    Args:
      uri (str): Uri.
      args (list): Args.
      kwargs (dict): Dictionary of params.
    Returns:
      Response (requests.Response): rsp.
    """
    url = self._make_url(uri, *args)
    return self._send('get', url, **kwargs)

  def post(self, uri, **kwargs):
    """
    Send a Post request to the server.
    Body can be either the dict or passed as kwargs
    headers is a dict.
    Args:
      uri (str): Uri.
      kwargs (dict): Dictionary of params.
    Returns:
      Response (requests.Response): rsp.
    """
    url = self._make_url(uri)
    return self._send('post', url, **kwargs)
    
  def make_rest_call(self, method, uri, *args, **kwargs):
    """This method calls the appropriate rest method based on the arguments.

    Args:
      method (str): HTTP method.
      uri (str): Relative_uri.
      args(any): Arguments.
      kwargs(dict): Key value pair for the additional args.

    Returns:
      rsp (dict): The response content loaded as a JSON.
    """
    func = getattr(self, method)
    return func(uri, *args, **kwargs)

  def _send(self, method, url, **kwargs):
    """This private method acting as proxy for all http methods.
    Args:
      method (str): The http method type.
      url (str): The URL to for the Request
      kwargs (dict): Keyword args to be passed to the requests call.
        retries (int): The retry count in case of HTTP errors.
                       Except the codes in the list NO_RETRY_HTTP_CODES.

    Returns:
      Response (requests.Response): The response object.
    """
    kwargs['verify'] = kwargs.get('verify', False)
    if 'timeout' not in kwargs:
      kwargs['timeout'] = self._timeout
    if 'data' not in kwargs:
      body = {}
      kwargs['data'] = json.dumps(body)
    content_dict = {'content-type': 'application/json'}
    kwargs.setdefault('headers', {})
    kwargs['headers'].update(content_dict)

    func = getattr(self._session, method)
    response = None

    retries = kwargs.pop("retries", None)
    retry_interval = kwargs.pop("retry_interval", self._retry_interval)
    retry_count = retries if retries else self._retries
    for ii in range(retry_count):
      try:
        response = func(url, **kwargs)
        if self._internal_debug:
          self._logger.debug("%s method The request url sent: %s" % (
            method.upper(), response.request.url))
          self._logger.debug('Response status: %d' % response.status_code)
          self._logger.debug('Response: %s' % json.dumps(response.json(),
                                                         indent=4))

      except (ConnectionError, ReadTimeout) as e:
        self._logger.warning("Request failed with error: %s" % e)
        if ii != retry_count - 1:
          time.sleep(retry_interval)
        continue
      finally:
        self._session.close()
      if response.ok:
        return response
      if response.status_code in [401, 403]:
        raise virt.VirtError('HTTP Auth Failed %s %s. \n res: response: %s' %
                             (method, url, response))
      elif response.status_code == 409:
        raise virt.VirtError('HTTP conflict with the current state of the '
                             'target resource %s %s. \n res: %s' %
                             (method, url, response))
      elif response.status_code in self.NO_RETRY_HTTP_CODES:
        break
      if ii != retry_count - 1:
        time.sleep(retry_interval)

    if response is not None:
      msg = 'HTTP %s %s failed: ' % (method, url)
      if hasattr(response, "text") and response.text:
        msg = "\n".join([msg, response.text]).encode('utf-8')
        raise virt.VirtError(msg)
    else:
      raise virt.VirtError("Failed to make the HTTP request (%s, %s)" %
                           (method, url))

  def get_tasks(self, timestamp, version, is_pc=False):
    """
    Returns a list of AHV tasks which happened after timestamp.
    Args:
      timestamp (int): Current timestamp.
      version (str): Interface version.
      is_pc (bool): Flag to determine f we need to poll for PC tasks.
    Returns:
      Task list (list): list of tasks.
    """
    ahv_clusters = self.get_ahv_cluster_uuids(version)
    (uri, cmd_method) = self.get_common_ver_url_and_method(cmd_key='get_tasks')
    # For task return. Use fv2.0 for now. update the url to use v2.0.
    url = self._url[:(self._url).rfind('v')] + 'v2.0' + uri

    res = self._send(method=cmd_method, url=url)
    data = res.json()

    if is_pc:
      return self.get_pc_tasks(data, timestamp, ahv_clusters)
    else:
      return self.get_pe_tasks(data, timestamp, ahv_clusters)

  def get_pc_tasks(self, data, timestamp, ahv_clusters):
    """
    Returns a list of AHV tasks on PC which happened after timestamp.
    Args:
      data (json): Rest response in json format.
      timestamp (str): Current timestamp.
      ahv_clusters (list): List of ahv clusters uuid.
    Returns:
      task_list (list): list of tasks on PC.
    """
    (uri, cmd_method) = self.get_common_ver_url_and_method(cmd_key='get_task')
    # For task return. Use fv2.0 for now. update the url to use v2.0.
    url = self._url[:(self._url).rfind('v')] + 'v2.0' + uri

    task_completed = False
    task_list = []
    if 'entities' in data:
      for task in data['entities']:
        if 'start_time_usecs' in task:
          if task['start_time_usecs'] > timestamp:

            if 'progress_status' in task:
              if task['progress_status'] in ahv_constants.TASK_COMPLETE_MSG:
                task_completed = True
            elif 'status' in task:
              if task['status'] in ahv_constants.TASK_COMPLETE_MSG:
                task_completed = True

            if task_completed:
              task_completed=False
              if 'subtask_uuid_list' in task:
                for subtask in task['subtask_uuid_list']:
                  url = url % subtask
                  subtask_resp = self._send(cmd_method, url)
                  subtask_data = subtask_resp.json()

                  if 'progress_status' in subtask_data:
                    if subtask_data['progress_status'] in \
                      ahv_constants.TASK_COMPLETE_MSG:

                      if 'cluster_uuid' in subtask_data:
                        cluster_uuid = subtask_data['cluster_uuid']
                      else:
                        # Task does not have any cluster associated with it,
                        # skip it.
                        continue

                      if cluster_uuid in ahv_clusters:
                        if 'entity_list' in task:
                          entity_type_list = task['entity_list']
                        else:
                          # Task doesn't have any entity list, skip it.
                          continue

                      if entity_type_list:
                        for ent_type in entity_type_list:
                          if 'entity_type' in ent_type:
                            if (str(ent_type['entity_type'])).lower() \
                                    in self.event_types:
                              task_list.append(task)
                              task_list.append(subtask_data)

                    else:
                      # Task has not finished or it failed, skip it and continue
                      # the loop
                      continue

    return task_list



  def get_pe_tasks(self, data, timestamp, ahv_clusters):
    """
    Returns a list of AHV tasks on PE which happened after timestamp.
    Args:
      data (json): rest response in json format.
      timestamp (str): Current timestamp.
      ahv_clusters (list): list of ahv clusters uuid.
    Returns:
      task_list (list): list of tasks on PE.
    """
    task_completed = False
    task_list = []

    if 'entities' in data:
      for task in data['entities']:
        if 'start_time_usecs' in task:
          if task['start_time_usecs'] > timestamp:

            if 'progress_status' in task:
              if task['progress_status'] in ahv_constants.TASK_COMPLETE_MSG:
                task_completed = True
            elif 'status' in task:
              if task['status'] in ahv_constants.TASK_COMPLETE_MSG:
                task_completed = True

            if task_completed:
              task_completed = False
              if 'cluster_reference' in task:
                if 'uuid' in task['cluster_reference']:
                  cluster_uuid = task['cluster_reference']['uuid']
              elif 'cluster_uuid' in task:
                cluster_uuid = task['cluster_uuid']
              else:
                # Task does not have any cluster associated with it, skip it.
                continue

              if cluster_uuid in ahv_clusters:
                if 'entity_list' in task:
                  entity_type_list = task['entity_list']
                elif 'entity_reference_list' in task:
                  entity_type_list = task['entity_reference_list']
                else:
                  # Task doesn't have any entity list, skip it.
                  continue

                for ent_type in entity_type_list:
                  if 'entity_type' in ent_type:
                    if (str(ent_type['entity_type'])).lower() \
                            in self.event_types:
                      task_list.append(task)
                  elif 'kind' in ent_type:
                    if (str(ent_type['kind'])).lower() in self.event_types:
                      task_list.append(task)
                  else:
                    # Task doesn't have any event type associated to it.
                    continue
    return task_list

  def get_vms_uuid(self, version):
    """
    Returns the list of vms uuid.
    Args:
      version (str): Interface version.
    Returns:
      vm_uuid_list (list): list of vm's uuid.
    """
    vm_uuid_list = []
    (url, cmd_method) = self.get_diff_ver_url_and_method(
      cmd_key='list_vms', intf_version=version)
    res = self.make_rest_call(method=cmd_method, uri=url)
    data = res.json()
    if 'entities' in data:
      for vm_entity in data['entities']:
        if 'metadata' in vm_entity:
          vm_uuid_list.append(vm_entity['metadata']['uuid'])
        elif 'uuid' in vm_entity:
          vm_uuid_list.append(vm_entity['uuid'])
        else:
          self._logger.warning("Cannot access the uuid for the vm %s. "
                               "vm object: %s" % (vm_entity['name'], 
                                                  vm_entity))
    return vm_uuid_list

  def get_hosts_uuid(self, version):
    """
    Returns the list of host uuid.
    Args:
      version (str): Interface version.
    Returns:
      host_uuid_list (list): list of host's uuid.
    """
    host_uuid_list = []
    (url, cmd_method) = self.get_diff_ver_url_and_method(
      cmd_key='list_hosts', intf_version=version)

    res = self.make_rest_call(method=cmd_method, uri=url)
    data = res.json()
    if 'entities' in data:
      for host_entity in data['entities']:
        if 'status' in host_entity and'metadata' in host_entity:
          # Check if a physical host, not a cluster.
          if 'cpu_model' in host_entity['status']:
            host_uuid_list.append(host_entity['metadata']['uuid'])
        elif 'uuid' in host_entity:
          host_uuid_list.append(host_uuid_list['uuid'])
        else:
          self._logger.warning("Cannot access the uuid for the. "
                               "host object: %s" % (host_entity))


  def get_host_cluster_uuid(self, host_info):
    """
    Returns host's cluster UUID.
    Args:
      host_info (dict): Host info dict.
    Returns:
      host_cluster_uuid (uuid): host's cluster uuid.
    """
    if 'cluster_uuid' in host_info:
      return host_info['cluster_uuid']
    elif 'cluster_reference' in host_info:
      return host_info['cluster_reference']['uuid']

  def get_ahv_cluster_uuids(self, version):
    """
    Returns list of ahv cluster uuids.
    Args:
      version (str): Interface version.
    Returns:
      ahv_host_cluster_uuids (List): Returns list of ahv cluster uuids.
    """
    ahv_host_cluster_uuids = []
    seen = set(ahv_host_cluster_uuids)

    (url, cmd_method) = self.get_diff_ver_url_and_method(
      cmd_key='list_clusters', intf_version=version)
    res = self.make_rest_call(method=cmd_method, uri=url)
    data = res.json()

    formatted_data = self._format_response(data)

    for cluster in formatted_data:
      if 'hypervisor_types' in cluster and 'cluster_uuid' in cluster:
        for hypevirsor_type in cluster['hypervisor_types']:
          if hypevirsor_type in ahv_constants.AHV_HYPERVIRSOR:
            cluster_uuid = cluster['cluster_uuid']
            if cluster_uuid not in seen:
              seen.add(cluster_uuid)
              ahv_host_cluster_uuids.append(cluster['cluster_uuid'])
              break

    return ahv_host_cluster_uuids

  def get_host_version(self, host_info):
    """
    Returns host's version.
    Args:
      host_info (dict): Host info dict.
    Returns:
      host_version (Str): Host version if found, None otherwise.
    """
    host_version = None
    if 'resources' in host_info:
      host_resources = host_info['resources']
      if 'hypervisor' in host_resources:
        if 'hypervisor_full_name' in host_resources['hypervisor']:
          host_version = host_resources['hypervisor']['hypervisor_full_name']
    elif 'hypervisor_full_name' in host_info:
      host_version = host_info['hypervisor_full_name']
    else:
      self._logger.warning("Cannot get host version for %s"
                                % host_info['uuid'])
    
    return host_version

  def get_vm(self, uuid):
    """
    Returns vm information
    Args:
      uuid (str): Vm uuid.
    Return:
      data (dict): Vm information.
    """
    (url, cmd_method) = self.get_common_ver_url_and_method(cmd_key='get_vm')
    url = url % uuid
    res = self.make_rest_call(method=cmd_method, uri=url)
    data = res.json()

    return self._format_response(data)

  def get_host(self, uuid):
    """
    Returns host information
    Args:
      uuid (str): Host uuid.
    Return:
      data (dict): Host information.
    """
    (url, cmd_method) = self.get_common_ver_url_and_method(cmd_key='get_host')
    url = url % uuid
    res = self.make_rest_call(method=cmd_method, uri=url)
    data = res.json()

    return self._format_response(data)

  def get_vm_host_uuid_from_vm(self, vm_entity):
    """
    Get the host uuid from the vm_entity response
    Args:
      vm_entity (dict): Vm info.
    Returns:
      host uuid (str): Vm host uuid if found, none otherwise.
    """
    if 'resources' in vm_entity:
      if 'host_reference' in vm_entity['resources']:
        return vm_entity['resources']['host_reference']['uuid']
      else:
        self._logger.warning("Did not find any host information for vm:%s"
                            % vm_entity['uuid'])
    elif 'host_uuid' in vm_entity:
      return vm_entity['host_uuid']
    else:
      # Vm is off therefore no host is assigned to it.
      self._logger.debug('Cannot get the host uuid of the vm:%s. '
                         'perhaps the vm is powered off' % vm_entity['uuid'])
    return None

  def is_ahv_host(self, version, host_uuid, vm_entity=None):
    """
    Determine if a given host is a AHV host. 
    host uuid should match the host uuid in vm_entity.
    Args:
      version (str): API version.
      host_uuid (str): uuid of a host.
      vm_entity (dict): For v3
    Returns:
      bool : True if host is ahv; false otehrwise.
    """
    if version == 'v2.0':
      host = self.get_host(host_uuid)
      if 'hypervisor_type' in host:
        return host['hypervisor_type'] in ahv_constants.AHV_HYPERVIRSOR
    else:
      if 'resources' in vm_entity:
        if 'hypervisor_type' in vm_entity['resources']:
          return vm_entity['resources']['hypervisor_type'] in \
                 ahv_constants.AHV_HYPERVIRSOR
    self._logger.debug('Hypervisor type not found. \nversion:%s, '
                        '\nhost_uuid:%s, \nvm_entity:%s' 
                        % (version, host_uuid, vm_entity))
    return False

  def build_host_to_uvm_map(self, version):
    """
    Builds a dictionary of every ahv host along with the vms they are hosting
    Args:
      version (Str): API version
    Returns:
      host_uvm_map (dict): Dict of ahv host with its uvms.
    """
    host_uvm_map = {}
    vm_uuids =  self.get_vms_uuid(version)

    if len(vm_uuids) > 0:
      for vm_uuid in vm_uuids:
        vm_entity = self.get_vm(vm_uuid)
        host_uuid = self.get_vm_host_uuid_from_vm(vm_entity)
        if host_uuid:
          if self.is_ahv_host(version, host_uuid, vm_entity):
            host = self.get_host(host_uuid)
            if host_uuid not in host_uvm_map:
              host_uvm_map[host_uuid] = host
            if 'guest_list' in host_uvm_map[host_uuid]:
              host_uvm_map[host_uuid]['guest_list'].append(vm_entity)
            else:
              host_uvm_map[host_uuid]['guest_list'] = []
              host_uvm_map[host_uuid]['guest_list'].append(vm_entity)
          else:
            self._logger.debug("Host %s is not ahv, skipping it." % host_uuid)
            continue
          host_type = self.get_hypervisor_type(version, host, vm_entity)
          host_uvm_map[host_uuid]['hypervisor_type'] = host_type
    else:
      self._logger.warning("No available vms found")
      try:
        host_uuids = self.get_hosts_uuid(version)
        if len(host_uuids) > 0:
          for host_uuid in host_uuids:
            host = self.get_host(host_uuid)
            if host_uuid not in host_uvm_map:
              host_uvm_map[host_uuid] = host
              host_uvm_map[host_uuid]['guest_list'] = []

        else:
          self._logger.warning("No Available AHV host found")
      except TypeError:
        # In case there is no cluster registered to the PC.
        self._logger.warning("Unable to find any AHV hosts.")

    return host_uvm_map

class Failure(Exception):

  def __init__(self, details):
    self.details = details

  def __str__(self):
    try:
      return str(self.details)
    except Exception as exn:
      import sys
      print(exn)
      return "AHV-API failure: %s" % str(self.details)

  def _details_map(self):
    return dict([(str(i), self.details[i]) for i in range(len(self.details))])



# -*- coding: utf-8 -*-
from __future__ import print_function
"""
Module for logging, part of virt-who

Copyright (C) 2011 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""


import logging
import logging.handlers
import traceback
import os
import sys
import json
from six import StringIO
from six.moves.queue import Empty, Queue
from threading import Thread

from virtwho import util

try:
    from systemd import journal
except ImportError:
    journal = None

FILE_LOG_FORMAT = """%(asctime)s [%(levelname)s] @%(filename)s:%(lineno)d - %(message)s"""
STREAM_LOG_FORMAT = """%(asctime)s %(levelname)s: %(message)s"""
JOURNAL_LOG_FORMAT = """[%(levelname)s] @%(filename)s:%(lineno)d - %(message)s"""

DEBUG_FORMAT = "%(asctime)s [%(name)s %(levelname)s] " \
               "%(processName)s(%(process)d):%(threadName)s " \
               "@%(filename)s:%(funcName)s:%(lineno)d - %(message)s"

DEFAULT_FORMAT = DEBUG_FORMAT
DEFAULT_LOG_DIR = '/var/log/rhsm'
DEFAULT_LOG_FILE = 'rhsm.log'
DEFAULT_NAME = 'virtwho'


class QueueHandler(logging.Handler):
    """
    A handler that will write logrecords to a queue (Queue or multiprocessing.Queue)
    For use in logging in situations involving multiple processes
    """
    def __init__(self, queue, level=logging.NOTSET):
        logging.Handler.__init__(self, level)
        self._queue = queue
        self._formatter = self.formatter or logging.Formatter(FILE_LOG_FORMAT if level != logging.DEBUG else DEBUG_FORMAT)

    def formatException(self, ei):
        if self.level != logging.DEBUG:
            s = traceback.format_exception_only(ei[0], ei[1])[0]
        else:
            sio = StringIO()
            traceback.print_exception(ei[0], ei[1], ei[2], None, sio)
            s = sio.getvalue()
            sio.close()
        if s[-1:] == "\n":
            s = s[:-1]
        return s

    def prepare(self, record):
        """
        Prepares the record to be placed on the queue
        """
        if record.exc_info:
            record.exc_text = self.formatException(record.exc_info)
            record.exc_info = None

        # Apply string formatting to the message using args
        if hasattr(record, 'args'):
            record.msg = record.msg % record.args
            record.args = []

        serialized_record = json.dumps(record.__dict__)
        return serialized_record

    def emit(self, record):
        try:
            self._queue.put_nowait(self.prepare(record))
        except Exception:
            self.handleError(record)


class QueueLogger(object):
    """
    A threaded logger reading objects off a queue
    """
    def __init__(self, name, queue=None):
        self.name = name
        self.logger = logging.getLogger(self.name)
        self.queue = queue or Queue()
        self._logging_thread = Thread(target=QueueLogger._log,
                                      args=(self.logger,
                                            self.queue))

    @staticmethod
    def _log(logger, queue):
        want_exit = False
        while not want_exit:
            try:
                record = queue.get()
            except Empty:
                return
            if record:
                to_log = QueueLogger.prepare(record)
                if to_log:
                    logger.handle(to_log)
            else:
                want_exit = True

    def start_logging(self):
        self._logging_thread.start()

    def terminate(self):
        self.queue.put_nowait(None)
        if self._logging_thread:
            self._logging_thread.join()
            self._logging_thread = None

    @staticmethod
    def prepare(record):
        prepared_record = None
        try:
            deserialized_record = json.loads(record, object_hook=util.decode)
            prepared_record = logging.makeLogRecord(deserialized_record)
        except Exception:
            # Swallow exceptions
            pass
        return prepared_record

    def getHandler(self, level=logging.NOTSET):
        # Return a queue handler that will write to this queue logger
        return QueueHandler(self.queue, level)

    def addHandler(self, handler):
        self.logger.addHandler(handler)


def checkDir(directory):
    try:
        if not os.path.isdir(directory):
            os.mkdir(directory)
    except Exception as e:
        sys.stderr.write("Unable to create %s directory: %s" % (directory, str(e)))
        return False
    return True


class Logger(object):
    _log_dir = DEFAULT_LOG_DIR
    _log_file = DEFAULT_LOG_FILE
    _log_per_config = False
    _logs = {}
    _stream_handler = None
    _journal_handler = None
    _rhsm_file_handler = None
    _level = logging.DEBUG
    _rhsm_level = logging.WARN
    _queue_logger = None

    @classmethod
    def initialize(cls, log_dir=None, log_file=None, log_per_config=None, debug=None):
        # Set defaults if necessary
        if log_dir:
            cls._log_dir = log_dir
        if log_file:
            cls._log_file = log_file
        if log_per_config:
            cls._log_per_config = True
        cls._level = logging.DEBUG if debug else logging.INFO
        # We don't want INFO message from RHSM in non-debug mode
        cls._rhsm_level = logging.DEBUG if debug else logging.WARN

    @classmethod
    def get_logger(cls, name=None, config=None, queue=True):
        if name is None:
            # Remove slashes and periods in the log_file (as that could mess logging up
            name = config['global']['log_file'].replace('.', '_').replace('/', '_') if config else 'main'
        virt_who_logger_name = 'virtwho.' + name  # The name of the logger instance

        try:
            # Try to get an existing log
            return cls._logs[virt_who_logger_name]
        except KeyError:
            pass

        try:
            return cls._logs["rhsm"]
        except KeyError:
            pass

        logger = logging.getLogger(virt_who_logger_name)
        cls._logs[virt_who_logger_name] = logger
        # Because we are using or own queue logger we don't want any of the loggers to propagate
        logger.propagate = False
        logger.setLevel(cls._level)

        # Show logging from RHSM in the log when DEBUG is enabled
        rhsm_logger = logging.getLogger("rhsm")
        rhsm_logger.setLevel(cls._rhsm_level)

        rhsm_file_handler = cls.get_file_handler(name="rhsm", config=config)
        virt_who_file_handler = cls.get_file_handler(name=virt_who_logger_name, config=config)

        ppid = os.getppid()

        journal_handler = None
        stream_handler = None
        if ppid == 1:
            # we're running under systemd, log to journal
            journal_handler = cls.get_journal_handler()
        else:
            # we're not running under systemd, set up streamHandler
            stream_handler = cls.get_stream_handler(name)

        if queue:
            queue_logger = cls.get_queue_logger()
            # get a QueueHandler that will send to this queuelogger
            queue_handler = queue_logger.getHandler(cls._level)
            logger.addHandler(queue_handler)
            main_logger = queue_logger
        else:
            main_logger = logger

        if virt_who_file_handler:
            main_logger.addHandler(virt_who_file_handler)

        # We need only one file handler for log from RHSM
        if rhsm_file_handler and cls._rhsm_file_handler is None:
            rhsm_logger.addHandler(rhsm_file_handler)
            cls._rhsm_file_handler = rhsm_file_handler

        if stream_handler:
            main_logger.addHandler(stream_handler)
            rhsm_logger.addHandler(stream_handler)

        if journal_handler:
            main_logger.addHandler(journal_handler)
            rhsm_logger.addHandler(journal_handler)

        return logger

    @classmethod
    def get_file_handler(cls, name, config=None):
        if cls._log_per_config:
            log_file = name + '.log'
        else:
            log_file = cls._log_file

        checkDir(cls._log_dir)
        path = os.path.join(cls._log_dir, log_file)

        try:
            file_handler = logging.handlers.WatchedFileHandler(path)
        except Exception as e:
            sys.stderr.write("Unable to log to %s: %s\n" % (path, e))
            return None

        file_handler.addFilter(logging.Filter(name))
        file_handler.setLevel(cls._level)
        file_handler.setFormatter(logging.Formatter(FILE_LOG_FORMAT if cls._level != logging.DEBUG else DEBUG_FORMAT))
        return file_handler

    @classmethod
    def get_stream_handler(cls, name):
        if cls._stream_handler is not None:
            return cls._stream_handler
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(cls._level)
        stream_handler.setFormatter(logging.Formatter(STREAM_LOG_FORMAT if cls._level != logging.DEBUG else DEBUG_FORMAT))
        f = logging.Filter()
        f.filter = lambda record: record.exc_info is None
        stream_handler.addFilter(f)
        cls._stream_handler = stream_handler
        return stream_handler

    @classmethod
    def get_journal_handler(cls):
        if cls._journal_handler is None and journal:
            cls._journal_handler = journal.JournalHandler()
            cls._journal_handler.setLevel(cls._level)
            cls._journal_handler.setFormatter(logging.Formatter(JOURNAL_LOG_FORMAT if cls._level != logging.DEBUG else DEBUG_FORMAT))
        return cls._journal_handler

    @classmethod
    def get_queue_logger(cls):
        if cls._queue_logger is None:
            cls._queue_logger = QueueLogger(DEFAULT_NAME)
            cls._queue_logger.start_logging()
        return cls._queue_logger

    @classmethod
    def has_queue_logger(cls):
        return cls._queue_logger is not None


def init(config):
    log_dir = config['global']['log_dir']
    log_file = config['global']['log_file']
    log_per_config = config['global']['log_per_config']
    debug = config['global']['debug']
    return Logger.initialize(log_dir=log_dir, log_file=log_file, log_per_config=log_per_config,
                             debug=debug)


def getLogger(name=None, config=None, queue=True):
    """
    This method does the setup necessary to create and connect both
    the main logger instance used from virtwho as well as loggers for
    all the connected virt backends

    First the logger is created without the queue and it's added later on
    (after the fork).
    """

    return Logger.get_logger(name=name, config=config, queue=queue)


def getQueueLogger():
    return Logger.get_queue_logger()


def hasQueueLogger():
    return Logger.has_queue_logger()


def closeLogger(logger):
    while len(logger.handlers):
        h = logger.handlers[0]
        h.close()
        logger.removeHandler(h)



# -*- coding: utf-8 -*-
from __future__ import print_function

#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""
Test validating of LibvirtdConfigSection
"""

from base import TestBase
from mock import MagicMock
import tempfile
import os
import six
from binascii import hexlify
import shutil

from virtwho.password import Password
from virtwho.config import EffectiveConfig, ConfigSection, parse_file
from virtwho.virt.libvirtd.libvirtd import LibvirtdConfigSection

MY_SECTION_NAME = 'test-libvirt'
MY_SAT5_SECTION_NAME = 'test-sat5-libvirt'

# Values used for testing VirtConfigSection
LIBVIRT_SECTION_VALUES = {
    'type': 'libvirt',
    'server': 'ssh://10.0.0.101/system',
    'owner': '123456',
    'hypervisor_id': 'uuid',
    'filter_hosts': '*.example.com'
}

LIBVIRT_SECTION_TEXT = """
[test-libvirt]
type = libvirt
server = ssh://192.168.122.10
owner = 123456
"""

# No env or owner is required for Satellite 5
LIBVIRT_SAT5_SECTION_TEXT = """
[test-sat5-libvirt]
type = libvirt
sm_type = satellite
sat_server = https://sat5.company.com/XMLRPC
sat_username = admin
sat_password = secret
"""


class TestLibvirtdConfigSection(TestBase):
    """
    Test base for testing class LibvirtdConfigSection
    """

    def __init__(self, *args, **kwargs):
        super(TestLibvirtdConfigSection, self).__init__(*args, **kwargs)
        self.virt_config = None

    def init_virt_config_section(self):
        """
        Method executed before each unit test
        """
        self.virt_config = LibvirtdConfigSection(MY_SECTION_NAME, None)
        # We need to set values using this way, because we need
        # to trigger __setitem__ of virt_config
        for key, value in LIBVIRT_SECTION_VALUES.items():
            self.virt_config[key] = value

    def test_validate_libvirt_section(self):
        """
        Test validation of libvirt section
        """
        self.init_virt_config_section()
        result = self.virt_config.validate()
        self.assertGreater(len(result), 0)

    def test_validate_server_optional(self):
        """
        Test validation of server option. It is optional for libvirtd
        """
        self.init_virt_config_section()
        # The server option is set
        result = self.virt_config._validate_server('server')
        self.assertIsNone(result)
        # The server option is not set
        del self.virt_config['server']
        result = self.virt_config._validate_server('server')
        self.assertIsNone(result)

    def test_validate_server_full_url(self):
        """
        Test validation of libvirt URL containing all parts
        """
        self.init_virt_config_section()
        # Set full URL
        self.virt_config['server'] = 'qemu+ssh://admin@example.com/system'
        result = self.virt_config._validate_server('server')
        self.assertIsNone(result)

    def test_validate_uuid_hypervisor_id(self):
        """
        Test validation of valid hypervisor_id
        """
        # Following method sets hypervisor_id to 'uuid' and it is valid
        self.init_virt_config_section()
        result = self.virt_config._validate_hypervisor_id('hypervisor_id')
        self.assertIsNone(result)

    def test_validate_hostname_hypervisor_id(self):
        """
        Test validation of valid hypervisor_id
        """
        self.init_virt_config_section()
        self.virt_config['hypervisor_id'] = 'hostname'  # it is also valid value
        result = self.virt_config._validate_hypervisor_id('hypervisor_id')
        self.assertIsNone(result)

    def test_validate_unvalid_hypervisor_id(self):
        """
        Test validation of valid hypervisor_id
        """
        self.init_virt_config_section()
        self.virt_config['hypervisor_id'] = 'unsupported_id'
        result = self.virt_config._validate_hypervisor_id('hypervisor_id')
        expected_result = [
            ('error', 'Invalid option: "unsupported_id" for hypervisor_id, use one of: (uuid, hostname)')
        ]
        six.assertCountEqual(self, result, expected_result)

    def test_validate_server_url_missing_path(self):
        """
        Test validation of libvirt URL with missing path
        """
        self.init_virt_config_section()
        # Set full URL
        self.virt_config['server'] = 'qemu+ssh://admin@example.com'
        result = self.virt_config._validate_server('server')
        expected_result = [
            ('info', 'Libvirt path is not specified in the url, using /system')
        ]
        six.assertCountEqual(self, result, expected_result)
        self.assertEqual(self.virt_config['server'], 'qemu+ssh://admin@example.com/system?no_tty=1')

    def test_validate_server_url_missing_scheme(self):
        """
        Test validation of libvirt URL with missing scheme
        """
        self.init_virt_config_section()
        # Set full URL
        self.virt_config['server'] = 'example.com/system'
        result = self.virt_config._validate_server('server')
        expected_result = [
            ('info', 'Protocol is not specified in libvirt url, using qemu+ssh://')
        ]
        six.assertCountEqual(self, result, expected_result)
        self.assertEqual(self.virt_config['server'], 'qemu+ssh://example.com/system?no_tty=1')

    def test_validate_server_url_missing_scheme_and_path(self):
        """
        Test validation of libvirt URL with missing scheme and path
        """
        self.init_virt_config_section()
        # Set full URL
        self.virt_config['server'] = 'example.com'
        result = self.virt_config._validate_server('server')
        expected_result = [
            ('info', 'Protocol is not specified in libvirt url, using qemu+ssh://'),
            ('info', 'Libvirt path is not specified in the url, using /system')
        ]
        six.assertCountEqual(self, result, expected_result)
        self.assertEqual(self.virt_config['server'], 'qemu+ssh://example.com/system?no_tty=1')

    def test_validate_sam_owner(self):
        """
        Test validation of owner option for libvirtd virt. backend and SAM destination
        """
        self.init_virt_config_section()
        # When server is set and SAM destination is used, then owner has to be set
        assert 'server' in self.virt_config
        assert 'sam' == self.virt_config['sm_type']
        result = self.virt_config._validate_owner('owner')
        self.assertIsNone(result)
        # Delete 'owner' section
        del self.virt_config['owner']
        result = self.virt_config._validate_owner('owner')
        self.assertIsNotNone(result)
        # Delete server too, then owner need not to be set
        del self.virt_config['server']
        result = self.virt_config._validate_owner('owner')
        self.assertIsNone(result)

    def test_validate_sat5_owner(self):
        """
        Test validation of owner option for libvirtd virt. backend and SAT5 destination
        """
        self.init_virt_config_section()
        self.virt_config['sm_type'] = 'satellite'
        # When server is set and SAM destination is used, then owner has to be set
        assert 'server' in self.virt_config
        # Delete 'owner' section
        del self.virt_config['owner']
        result = self.virt_config._validate_owner('owner')
        self.assertIsNone(result)

    def test_validate_unencrypted_password(self):
        """
        test validation of unecrypted password option for libvirt virt. backend
        """
        self.init_virt_config_section()
        # Validation with no password option
        result = self.virt_config._validate_unencrypted_password('password')
        self.assertIsNone(result)
        # RHSM password or RHSM proxy can be set
        self.virt_config['rhsm_password'] = 'secret_password'
        result = self.virt_config._validate_unencrypted_password('rhsm_password')
        self.assertIsNone(result)
        self.virt_config['rhsm_proxy_password'] = 'secret_password'
        result = self.virt_config._validate_unencrypted_password('rhsm_proxy_password')
        self.assertIsNone(result)
        # When server is set, then setting password is useless
        self.virt_config['password'] = 'another_secret_password'
        result = self.virt_config._validate_unencrypted_password('password')
        self.assertIsNotNone(result)
        # When server is not set, then setting password is useless too, but
        # reason is different
        del self.virt_config['server']
        result = self.virt_config._validate_unencrypted_password('password')
        self.assertIsNotNone(result)

    def mock_pwd_file(self):
        # Backup previous values
        self.old_key_file = Password.KEYFILE
        self.old_can_write = Password._can_write
        # Mock pwd file
        f, filename = tempfile.mkstemp()
        self.addCleanup(os.unlink, filename)
        Password.KEYFILE = filename
        Password._can_write = MagicMock(return_value=True)

    def unmock_pwd_file(self):
        # Restore pwd file from backup
        Password.KEYFILE = self.old_key_file
        Password._can_write = self.old_can_write

    def test_validate_encrypted_password(self):
        """
        test validation of encrypted password option for libvirt virt. backend
        """
        self.mock_pwd_file()
        self.init_virt_config_section()
        # Validation with no encrypted_password option
        result = self.virt_config._validate_encrypted_password('encrypted_password')
        self.assertIsNone(result)
        # RHSM password or RHSM proxy can be set
        self.virt_config['rhsm_encrypted_password'] = hexlify(Password.encrypt('secret_password'))
        result = self.virt_config._validate_encrypted_password('rhsm_encrypted_password')
        self.assertIsNone(result)
        self.virt_config['rhsm_encrypted_proxy_password'] = hexlify(Password.encrypt('secret_password'))
        result = self.virt_config._validate_encrypted_password('rhsm_encrypted_proxy_password')
        self.assertIsNone(result)
        # When server is set, then setting password is useless
        self.virt_config['encrypted_password'] = hexlify(Password.encrypt('another_secret_password'))
        result = self.virt_config._validate_encrypted_password('encrypted_password')
        self.assertIsNotNone(result)
        # When server is not set, then setting password is useless too, but
        # reason is different
        del self.virt_config['server']
        result = self.virt_config._validate_encrypted_password('encrypted_password')
        self.assertIsNotNone(result)
        self.unmock_pwd_file()


class TestLibvirtEffectiveConfig(TestBase):
    """
    Test reading config section from text using EffectiveConfig
    """

    def __init__(self, *args, **kwargs):
        super(TestLibvirtEffectiveConfig, self).__init__(*args, **kwargs)
        self.effective_config = None

    def init_effective_config(self):
        """
        This method is executed before each unit test
        """
        self.effective_config = EffectiveConfig()

    def test_read_sam_effective_config_from_file(self):
        config_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, config_dir)
        with open(os.path.join(config_dir, "test.conf"), "w") as f:
            f.write(LIBVIRT_SECTION_TEXT)
        conf = parse_file(os.path.join(config_dir, "test.conf"))
        self.init_effective_config()
        conf_values = conf.pop(MY_SECTION_NAME)
        self.effective_config[MY_SECTION_NAME] = ConfigSection.from_dict(
            conf_values,
            MY_SECTION_NAME,
            self.effective_config
        )
        self.assertEqual(type(self.effective_config[MY_SECTION_NAME]), LibvirtdConfigSection)
        self.assertEqual(self.effective_config[MY_SECTION_NAME]['server'], 'ssh://192.168.122.10')
        validate_messages = self.effective_config.validate()
        self.assertIsNotNone(validate_messages)
        del self.effective_config[MY_SECTION_NAME]

    def test_read_sat5_effective_config_from_file(self):
        config_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, config_dir)
        with open(os.path.join(config_dir, "test.conf"), "w") as f:
            f.write(LIBVIRT_SAT5_SECTION_TEXT)
        conf = parse_file(os.path.join(config_dir, "test.conf"))
        self.init_effective_config()
        conf_values = conf.pop(MY_SAT5_SECTION_NAME)
        self.effective_config[MY_SAT5_SECTION_NAME] = ConfigSection.from_dict(
            conf_values,
            MY_SAT5_SECTION_NAME,
            self.effective_config
        )
        self.assertEqual(type(self.effective_config[MY_SAT5_SECTION_NAME]), LibvirtdConfigSection)
        validate_messages = self.effective_config.validate()
        self.assertEqual(self.effective_config[MY_SAT5_SECTION_NAME]['sat_server'], 'https://sat5.company.com/XMLRPC')
        self.assertEqual(self.effective_config[MY_SAT5_SECTION_NAME]['sm_type'], 'satellite')
        self.assertIsNotNone(validate_messages)
        del self.effective_config[MY_SAT5_SECTION_NAME]



import os
import subprocess

from tito.distributionbuilder import DistributionBuilder
from tito.common import info_out

class PatchBuilder(DistributionBuilder):
    """Builder that includes a patch files in the tarball.
    """
    def __init__(self, config=None, *args, **kwargs):
        super(PatchBuilder, self).__init__(config=config, *args, **kwargs)


    def tgz(self):
        retval = DistributionBuilder.tgz(self)
        self.sources.append(os.path.join(self.rpmbuild_basedir, 'build-rpm-no-ahv.patch'))
        self.artifacts.append(os.path.join(self.rpmbuild_basedir, 'build-rpm-no-ahv.patch'))
        subprocess.check_call("cp %s %s/" % ('build-rpm-no-ahv.patch', self.rpmbuild_sourcedir), shell=True)

        return retval



from __future__ import print_function

import os
import tempfile
import shutil

from base import TestBase
from stubs import StubEffectiveConfig

from mock import Mock, patch, call
from threading import Event

from virtwho import MinimumJobPollInterval
from virtwho.config import DestinationToSourceMapper, VW_GLOBAL, EffectiveConfig, parse_file, \
    VirtConfigSection
from virtwho.manager import ManagerThrottleError
from virtwho.virt import HostGuestAssociationReport, Hypervisor, Guest, \
    DestinationThread, ErrorReport, AbstractVirtReport, DomainListReport


xvirt = type("", (), {'CONFIG_TYPE': 'xxx'})()


class TestVirtInclude(TestBase):
    def test_filter_hosts(self):
        self.filter_hosts('filter_hosts=12345')

    def test_exclude_hosts(self):
        self.filter_hosts('exclude_hosts=00000')

    def test_filter_hosts_glob(self):
        self.filter_hosts('filter_hosts=12*')
        self.filter_hosts('filter_hosts=12?45')
        self.filter_hosts('filter_hosts=12[36]45')

    def test_filter_hosts_glob_filter_type(self):
        self.filter_hosts('filter_hosts=12*', 'filter_type=wildcards')
        self.filter_hosts('filter_hosts=12?45', 'filter_type=wildcards')
        self.filter_hosts('filter_hosts=12[36]45', 'filter_type=wildcards')

    def test_exclude_hosts_glob(self):
        self.filter_hosts('exclude_hosts=00*')
        self.filter_hosts('exclude_hosts=00?00')
        self.filter_hosts('exclude_hosts=00[03]00')

    def test_filter_hosts_regex(self):
        self.filter_hosts('filter_hosts=12.*')
        self.filter_hosts('filter_hosts=12.+45')
        self.filter_hosts('filter_hosts=12[36]45')

    def test_filter_hosts_regex_filter_type(self):
        self.filter_hosts('filter_hosts=12.*', 'filter_type=regex')
        self.filter_hosts('filter_hosts=12.+45', 'filter_type=regex')
        self.filter_hosts('filter_hosts=12[36]45', 'filter_type=regex')

    def test_exclude_hosts_regex(self):
        self.filter_hosts('exclude_hosts=00.*')
        self.filter_hosts('exclude_hosts=00.+00')
        self.filter_hosts('exclude_hosts=00[03]00')

    def test_filter_host_uuids(self):
        self.filter_hosts('filter_host_uuids=12345')

    def test_exclude_host_uuids(self):
        self.filter_hosts('exclude_host_uuids=00000')

    def filter_hosts(self, filter_something, filter_type=''):
        config_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, config_dir)
        with open(os.path.join(config_dir, "test.conf"), "w") as f:
            f.write("""
[test]
type=esx
server=does.not.exist
username=username
password=password
owner=owner
{filter_something}
{filter_type}
""".format(filter_something=filter_something, filter_type=filter_type))
        conf = parse_file(os.path.join(config_dir, "test.conf"))
        test_conf_values = conf.pop('test')
        effective_config = EffectiveConfig()
        effective_config['test'] = VirtConfigSection.from_dict(
            test_conf_values,
            'test',
            effective_config
        )
        effective_config.validate()
        config_manager = DestinationToSourceMapper(effective_config)
        self.assertEqual(len(config_manager.configs), 1)
        config = config_manager.configs[0][1]

        included_hypervisor = Hypervisor('12345', guestIds=[
            Guest('guest-2', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING),
        ])
        excluded_hypervisor = Hypervisor('00000', guestIds=[
            Guest('guest-1', xvirt.CONFIG_TYPE, Guest.STATE_RUNNING),
        ])

        assoc = {
            'hypervisors': [
                excluded_hypervisor,
                included_hypervisor,
            ]
        }

        report = HostGuestAssociationReport(config, assoc)
        assert report.association == {
            'hypervisors': [
                included_hypervisor
            ]
        }


class TestDestinationThread(TestBase):

    default_config_args = {
        'type': 'esx',
        'hypervisor_id': 'uuid',
        'simplified_vim': True,
        'owner': 'owner',
    }

    def setUp(self):
        self.options_values = {
            VW_GLOBAL: {
                'print': False,
            },
        }
        self.options = StubEffectiveConfig(self.options_values)

    def test_get_data(self):
        # Show that get_data accesses the given source and tries to retrieve
        # the right source_keys
        source_keys = ['source1', 'source2']
        report1 = Mock()
        report2 = Mock()
        report1.hash = "report1_hash"
        report2.hash = "report2_hash"
        datastore = {'source1': report1, 'source2': report2}
        manager = Mock()
        logger = Mock()
        config, d = self.create_fake_config('test', **self.default_config_args)
        terminate_event = Mock()
        interval = 10  # Arbitrary for this test
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=True, options=self.options)
        destination_thread.is_initial_run = False
        result_data = destination_thread._get_data()
        self.assertEqual(result_data, datastore)

    def test_get_data_ignore_same_reports(self):
        # Show that the data returned from _get_data does not include those
        # reports whose hash is identical to that of the last one sent for
        # the given source
        source_keys = ['source1', 'source2']
        report1 = Mock()
        report2 = Mock()
        report1.hash = "report1_hash"
        report2.hash = "report2_hash"
        datastore = {'source1': report1, 'source2': report2}
        last_report_for_source = {
            'source1': report1.hash
        }
        expected_data = {
            'source2': report2
        }
        manager = Mock()
        logger = Mock()
        config, d = self.create_fake_config('test', **self.default_config_args)
        terminate_event = Mock()
        interval = 10  # Arbitrary for this test
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=True,
                                               options=self.options)
        destination_thread.is_initial_run = False
        destination_thread.last_report_for_source = last_report_for_source
        result_data = destination_thread._get_data()
        self.assertEqual(result_data, expected_data)

    @patch('virtwho.virt.virt.Event')
    def test_send_data_quit_on_error_report(self, mock_event_class):
        mock_event = Mock(spec=Event())
        mock_event_class.return_value = mock_event

        mock_error_report = Mock(spec=ErrorReport)

        source_keys = ['source1', 'source2']
        report1 = Mock()
        report2 = Mock()
        report1.hash = "report1_hash"
        report2.hash = "report2_hash"
        datastore = {'source1': report1, 'source2': report2}
        manager = Mock()
        logger = Mock()
        config, d = self.create_fake_config('test', **self.default_config_args)
        terminate_event = Mock()
        interval = 10  # Arbitrary for this test
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=True,
                                               options=self.options)
        destination_thread._send_data(mock_error_report)
        mock_event.set.assert_called()

    def test_send_data_batch_hypervisor_checkin(self):
        # This tests that reports of the right type are batched into one
        # and that the hypervisorCheckIn method of the destination is called
        # with the right parameters
        config1, d1 = self.create_fake_config('source1', **self.default_config_args)
        d1['exclude_hosts'] = []
        d1['filter_hosts'] = []

        config2, d2 = self.create_fake_config('source2', **self.default_config_args)
        d2['exclude_hosts'] = []
        d2['filter_hosts'] = []

        virt1 = Mock()
        virt1.CONFIG_TYPE = 'esx'
        virt2 = Mock()
        virt2.CONFIG_TYPE = 'esx'

        guest1 = Guest('GUUID1', virt1.CONFIG_TYPE, Guest.STATE_RUNNING)
        guest2 = Guest('GUUID2', virt2.CONFIG_TYPE, Guest.STATE_RUNNING)
        assoc1 = {'hypervisors': [Hypervisor('hypervisor_id_1', [guest1])]}
        assoc2 = {'hypervisors': [Hypervisor('hypervisor_id_2', [guest2])]}
        report1 = HostGuestAssociationReport(config1, assoc1)
        report2 = HostGuestAssociationReport(config2, assoc2)

        data_to_send = {'source1': report1,
                        'source2': report2}

        source_keys = ['source1', 'source2']
        report1 = Mock()
        report2 = Mock()
        report1.hash = "report1_hash"
        report2.hash = "report2_hash"
        datastore = {'source1': report1, 'source2': report2}
        manager = Mock()
        options = Mock()
        options.print_ = False

        def check_hypervisorCheckIn(report, options=None):
            self.assertEqual(report.association['hypervisors'],
                              data_to_send.values)

        manager.hypervisorCheckIn = Mock(side_effect=check_hypervisorCheckIn)
        logger = Mock()
        config, d = self.create_fake_config('test', **self.default_config_args)
        terminate_event = Mock()
        interval = 10  # Arbitrary for this test
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=True, options=self.options)
        destination_thread._send_data(data_to_send)

    def test_send_data_poll_hypervisor_async_result(self):
        # This test's that when we have an async result from the server,
        # we poll for the result

        # Setup the test data
        config1, d1 = self.create_fake_config('source1', **self.default_config_args)
        config2, d2 = self.create_fake_config('source2', **self.default_config_args)
        virt1 = Mock()
        virt1.CONFIG_TYPE = 'esx'
        virt2 = Mock()
        virt2.CONFIG_TYPE = 'esx'

        guest1 = Guest('GUUID1', virt1.CONFIG_TYPE, Guest.STATE_RUNNING)
        guest2 = Guest('GUUID2', virt2.CONFIG_TYPE, Guest.STATE_RUNNING)
        assoc1 = {'hypervisors': [Hypervisor('hypervisor_id_1', [guest1])]}
        assoc2 = {'hypervisors': [Hypervisor('hypervisor_id_2', [guest2])]}
        report1 = HostGuestAssociationReport(config1, assoc1)
        report2 = HostGuestAssociationReport(config2, assoc2)

        data_to_send = {'source1': report1,
                        'source2': report2}
        source_keys = ['source1', 'source2']
        batch_report1 = Mock()  # The "report" to check status
        batch_report1.state = AbstractVirtReport.STATE_CREATED
        datastore = {'source1': report1, 'source2': report2}
        manager = Mock()
        items = [ManagerThrottleError(), ManagerThrottleError(), ManagerThrottleError(), AbstractVirtReport.STATE_FINISHED]
        manager.check_report_state = Mock(side_effect=self.check_report_state_closure(items))

        logger = Mock()
        config, d = self.create_fake_config('test', **self.default_config_args)
        terminate_event = Mock()
        interval = 10  # Arbitrary for this test
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=False, options=self.options)
        # In this test we want to see that the wait method is called when we
        # expect and with what parameters we expect
        destination_thread.wait = Mock()
        destination_thread.is_terminated = Mock(return_value=False)
        destination_thread.check_report_status(batch_report1)
        # There should be three waits, one after the job is submitted with duration of
        # MinimumJobPollingInterval. The second and third with duration MinimumJobPollInterval * 2
        # (and all subsequent calls as demonstrated by the third wait)
        destination_thread.wait.assert_has_calls([
            call(wait_time=MinimumJobPollInterval),
            call(wait_time=MinimumJobPollInterval * 2),
            call(wait_time=MinimumJobPollInterval * 2)])

    def test_status_pending_hypervisor_async_result(self):
        # This test's that when we have an async result from the server,
        # we poll for the status on the interval until we get completed result

        # Setup the test data
        config1, d1 = self.create_fake_config('source1', **self.default_config_args)
        config2, d2 = self.create_fake_config('source2', **self.default_config_args)
        virt1 = Mock()
        virt1.CONFIG_TYPE = 'esx'
        virt2 = Mock()
        virt2.CONFIG_TYPE = 'esx'

        guest1 = Guest('GUUID1', virt1.CONFIG_TYPE, Guest.STATE_RUNNING)
        guest2 = Guest('GUUID2', virt2.CONFIG_TYPE, Guest.STATE_RUNNING)
        assoc1 = {'hypervisors': [Hypervisor('hypervisor_id_1', [guest1])]}
        assoc2 = {'hypervisors': [Hypervisor('hypervisor_id_2', [guest2])]}
        report1 = HostGuestAssociationReport(config1, assoc1)
        report2 = HostGuestAssociationReport(config2, assoc2)
        report1.job_id = 'job1'
        report2.job_id = 'job2'

        data_to_send = {'source1': report1,
                        'source2': report2}
        source_keys = ['source1', 'source2']
        batch_report1 = Mock()  # The "report" to check status
        batch_report1.state = AbstractVirtReport.STATE_CREATED
        datastore = {'source1': report1, 'source2': report2}
        manager = Mock()
        items = [AbstractVirtReport.STATE_PROCESSING, AbstractVirtReport.STATE_PROCESSING,
                 AbstractVirtReport.STATE_PROCESSING, AbstractVirtReport.STATE_FINISHED,
                 AbstractVirtReport.STATE_FINISHED]
        manager.check_report_state = Mock(side_effect=self.check_report_state_closure(items))

        logger = Mock()
        config, d = self.create_fake_config('test', **self.default_config_args)
        terminate_event = Mock()
        interval = 10  # Arbitrary for this test
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=False, options=self.options)
        # In this test we want to see that the wait method is called when we
        # expect and with what parameters we expect
        destination_thread.wait = Mock()
        destination_thread.is_terminated = Mock(return_value=False)
        destination_thread.submitted_report_and_hash_for_source ={'source1':(report1, 'hash1'),'source2':(report2, 'hash2')}
        reports = destination_thread._get_data_common(source_keys)
        self.assertEqual(0, len(reports))
        reports = destination_thread._get_data_common(source_keys)
        self.assertEqual(1, len(reports))
        reports = destination_thread._get_data_common(source_keys)
        self.assertEqual(2, len(reports))


    # A closure to allow us to have a function that "modifies" the given
    # report in a predictable way.
    # In this case I want to set the state of the report to STATE_FINISHED
    # after the first try

    def check_report_state_closure(self, items):
        item_iterator = iter(items)

        def mock_check_report_state(report):
            item = next(item_iterator)
            if isinstance(item, Exception):
                raise item
            report.state = item
            return report

        return mock_check_report_state

    def test_send_data_poll_async_429(self):
        # This test's that when a 429 is detected during async polling
        # we wait for the amount of time specified
        source_keys = ['source1', 'source2']
        config1, d1 = self.create_fake_config('source1', **self.default_config_args)
        config2, d2 = self.create_fake_config('source2', **self.default_config_args)
        virt1 = Mock()
        virt1.CONFIG_TYPE = 'esx'
        virt2 = Mock()
        virt2.CONFIG_TYPE = 'esx'

        guest1 = Guest('GUUID1', virt1.CONFIG_TYPE, Guest.STATE_RUNNING)
        guest2 = Guest('GUUID2', virt2.CONFIG_TYPE, Guest.STATE_RUNNING)
        assoc1 = {'hypervisors': [Hypervisor('hypervisor_id_1', [guest1])]}
        assoc2 = {'hypervisors': [Hypervisor('hypervisor_id_2', [guest2])]}
        report1 = HostGuestAssociationReport(config1, assoc1)
        report2 = HostGuestAssociationReport(config2, assoc2)

        datastore = {'source1': report1, 'source2': report2}
        data_to_send = {'source1': report1,
                        'source2': report2}
        config, d = self.create_fake_config('test', **self.default_config_args)
        error_to_throw = ManagerThrottleError(retry_after=62)

        manager = Mock()
        manager.hypervisorCheckIn = Mock(side_effect=[error_to_throw, report1])
        expected_wait_calls = [call(wait_time=error_to_throw.retry_after)]

        logger = Mock()
        terminate_event = Mock()
        interval = 10  # Arbitrary for this test
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=False, options=self.options)
        destination_thread.wait = Mock()
        destination_thread.is_terminated = Mock(return_value=False)
        destination_thread._send_data(data_to_send)
        destination_thread.wait.assert_has_calls(expected_wait_calls)

    def test_send_data_domain_list_reports(self):
        # Show that DomainListReports are sent using the sendVirtGuests
        # method of the destination

        source_keys = ['source1']
        config1, d1 = self.create_fake_config('source1', **self.default_config_args)
        virt1 = Mock()
        virt1.CONFIG_TYPE = 'esx'

        guest1 = Guest('GUUID1', virt1.CONFIG_TYPE, Guest.STATE_RUNNING)
        report1 = DomainListReport(config1, [guest1],
                                   hypervisor_id='hypervisor_id_1')

        datastore = {'source1': report1}
        data_to_send = {'source1': report1}

        config, d = self.create_fake_config('test', **self.default_config_args)
        logger = Mock()

        manager = Mock()
        terminate_event = Mock()
        interval = 10
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=True, options=self.options)
        destination_thread.wait = Mock()
        destination_thread.is_terminated = Mock(return_value=False)
        destination_thread._send_data(data_to_send)
        manager.sendVirtGuests.assert_has_calls([call(report1,
                                                      options=destination_thread.options)])

    def test_send_data_429_during_send_virt_guests(self):
        # Show that when a 429 is encountered during the sending of a
        # DomainListReport that we retry after waiting the appropriate
        # amount of time
        config, d = self.create_fake_config('test', **self.default_config_args)
        source_keys = ['source1']
        virt1 = Mock()
        virt1.CONFIG_TYPE = 'esx'

        guest1 = Guest('GUUID1', virt1.CONFIG_TYPE, Guest.STATE_RUNNING)
        report1 = DomainListReport(config, [guest1],
                                   hypervisor_id='hypervisor_id_1')

        datastore = {'source1': report1}
        data_to_send = {'source1': report1}

        logger = Mock()

        error_to_throw = ManagerThrottleError(retry_after=62)

        manager = Mock()
        manager.sendVirtGuests = Mock(side_effect=[error_to_throw, report1])
        terminate_event = Mock()
        interval = 10
        options = Mock()
        options.print_ = False
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=False, options=self.options)
        destination_thread.wait = Mock()
        destination_thread.is_terminated = Mock(return_value=False)
        destination_thread._send_data(data_to_send)
        manager.sendVirtGuests.assert_has_calls([call(report1,
                                                      options=destination_thread.options)])
        destination_thread.wait.assert_has_calls([call(
                wait_time=error_to_throw.retry_after)])

    def test_duplicate_reports_are_ignored(self):
        """
        Test that duplicate reports are filtered out when retrieving items
        from the data store
        """
        source_keys = ['source1', 'source2']
        interval = 1
        terminate_event = Mock()
        virt1 = Mock()
        virt1.CONFIG_TYPE = 'esx'
        config, d = self.create_fake_config('test', **self.default_config_args)
        manager = Mock()

        guest1 = Guest('GUUID1', virt1.CONFIG_TYPE, Guest.STATE_RUNNING)
        report1 = DomainListReport(config, [guest1],
                                   hypervisor_id='hypervisor_id_1')
        report2 = DomainListReport(config, [guest1],
                                   hypervisor_id='hypervisor_id_2')
        report3 = DomainListReport(config, [guest1],
                                   hypervisor_id='hypervisor_id_3')
        datastore = {
            'source1': report1,  # Not changing, should be excluded later
            'source2': report2,  # Will change the report sent for source2
        }
        data_to_send = {
            'source1': report1,
            'source2': report2,
        }
        logger = Mock()
        destination_thread = DestinationThread(logger, config,
                                               source_keys=source_keys,
                                               source=datastore,
                                               dest=manager,
                                               interval=interval,
                                               terminate_event=terminate_event,
                                               oneshot=False, options=self.options)
        destination_thread.is_initial_run = False
        destination_thread.is_terminated = Mock(return_value=False)
        destination_thread._send_data(data_to_send=data_to_send)

        expected_hashes = {}
        for source_key, report in data_to_send.items():
            expected_hashes[source_key] = report.hash

        self.assertEqual(destination_thread.last_report_for_source,
                         expected_hashes)
        # Pretend there were updates to the datastore from elsewhere
        destination_thread.source['source2'] = report3

        next_data_to_send = destination_thread._get_data()
        expected_next_data_to_send = {
            'source2': report3
        }
        self.assertEqual(next_data_to_send, expected_next_data_to_send)


class TestDestinationThreadTiming(TestBase):
    """
    A group of tests meant to show that the destination thread does things
    in the right amount of time given different circumstances.
    """
    pass



from __future__ import print_function

import six

from base import TestBase
from mock import patch, call, ANY, MagicMock
from requests import Session
from six.moves.queue import Queue
from threading import Event

from virtwho import DefaultInterval
from virtwho.datastore import Datastore
from virtwho.virt.ahv.ahv import AhvConfigSection
from virtwho.virt import Virt, VirtError, Guest, Hypervisor



MY_SECTION_NAME = 'test-ahv'
DefaultUpdateInterval = 1800
# Values used for testing AhvConfigSection.
PE_SECTION_VALUES = {
    'type': 'ahv',
    'server': '10.10.10.10',
    'username': 'root',
    'password': 'root_password',
    'owner': 'nutanix',
    'hypervisor_id': 'uuid',
    'is_hypervisor': True,
    'internal_debug': False,
    'update_interval': 60
}

HOST_UVM_MAP = \
 {u'08469de5-be42-43e6-8c32-20167d3b58f7':
           {u'oplog_disk_pct': 3.4,
            u'memory_capacity_in_bytes': 135009402880,
            u'has_csr': False,
            u'default_vm_storage_container_uuid': None,
            u'hypervisor_username': u'root',
            u'key_management_device_to_certificate_status': {},
            u'service_vmnat_ip': None,
            u'hypervisor_key': u'10.53.97.188',
            u'acropolis_connection_state': u'kConnected',
            u'management_server_name': u'10.53.97.188',
            u'failover_cluster_fqdn': None,
            u'serial': u'OM155S016008',
            u'bmc_version': u'01.92',
            u'hba_firmwares_list':
             [{u'hba_model': u'LSI Logic SAS3008',
               u'hba_version': u'MPTFW-06.00.00.00-IT'}],
            u'hypervisor_state': u'kAcropolisNormal',
            u'num_cpu_threads': 32,
            u'monitored': True,
            u'uuid': u'08469de5-be42-43e6-8c32-20167d3b58f7',
            u'reboot_pending': False,
            u'cpu_capacity_in_hz': 38384000000,
            u'num_cpu_sockets': 2,
            u'host_maintenance_mode_reason': None,
            u'hypervisor_address': u'10.53.97.188',
            u'host_gpus': None,
            u'failover_cluster_node_state': None,
            u'state': u'NORMAL',
            u'num_cpu_cores': 16,
            'guest_list':
             [{u'vm_features':
                {u'AGENT_VM': False,
                 u'VGA_CONSOLE': True},
               u'name': u'am2', u'num_cores_per_vcpu': 1,
               u'gpus_assigned': False, u'num_vcpus': 2,
               u'memory_mb': 4096,
               u'power_state': u'on',
               u'ha_priority': 0,
               u'allow_live_migrate': True,
               u'timezone': u'America/Los_Angeles',
               u'vm_logical_timestamp': 48,
               u'host_uuid': u'08469de5-be42-43e6-8c32-20167d3b58f7',
               u'uuid': u'01dcfc0b-3092-4f1b-94fb-81b44ed352be'},
              {u'vm_features':
                {u'AGENT_VM': False, u'VGA_CONSOLE': True},
               u'name': u'am3',
               u'num_cores_per_vcpu': 1,
               u'gpus_assigned': False,
               u'num_vcpus': 2, u'memory_mb': 4096,
               u'power_state': u'on',
               u'ha_priority': 0,
               u'allow_live_migrate': True,
               u'timezone': u'America/Los_Angeles',
               u'vm_logical_timestamp': 3,
               u'host_uuid': u'08469de5-be42-43e6-8c32-20167d3b58f7',
               u'uuid': u'422f9171-db1f-48b0-a3de-b0bb92a8f559'},
              {u'vm_features':
                {u'AGENT_VM': False,
                 u'VGA_CONSOLE': True},
               u'name': u'win_vm',
               u'num_cores_per_vcpu': 1,
               u'gpus_assigned': False,
               u'num_vcpus': 2,
               u'memory_mb': 4096,
               u'power_state': u'on',
               u'ha_priority': 0,
               u'allow_live_migrate': True,
               u'timezone': u'America/Los_Angeles',
               u'vm_logical_timestamp': 3,
               u'host_uuid': u'08469de5-be42-43e6-8c32-20167d3b58f7',
               u'uuid': u'98839f35-bd62-4255-a7cd-7668bc143554'}],
            u'cpu_model': u'Intel(R) Xeon(R) CPU E5-2630 v3 @ 2.40GHz',
            u'ipmi_username': u'ADMIN',
            u'service_vmid': u'0005809e-62e4-75c7-611b-0cc47ac3b354::7',
            u'bmc_model': u'X10_ATEN',
            u'host_nic_ids': [],
            u'cluster_uuid': u'0005809e-62e4-75c7-611b-0cc47ac3b354',
            u'ipmi_password': None,
            u'cpu_frequency_in_hz': 2399000000,
            u'stats':
             {u'num_read_io': u'8',
              u'controller_read_io_bandwidth_kBps': u'0',
              u'content_cache_hit_ppm': u'1000000',
              },
            u'num_vms': 4, u'default_vm_storage_container_id': None,
            u'metadata_store_status': u'kNormalMode',
            u'name': u'foyt-4', u'hypervisor_password': None,
            u'service_vmnat_port': None,
            u'hypervisor_full_name': u'Nutanix 20180802.100874',
            u'is_degraded': False, u'host_type': u'HYPER_CONVERGED',
            u'default_vhd_storage_container_uuid': None,
            u'block_serial': u'15SM60250038',
            u'disk_hardware_configs':
             {u'1':
               {
                u'mount_path': u'/home/nutanix/data/stargate-storage/disks/BTHC506101XL480MGN',
                },
              u'3':
               {
                u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8E6QE',
                },
              u'2':
               {
                u'mount_path': u'/home/nutanix/data/stargate-storage/disks/BTHC50610246480MGN',
                },
              u'5':
               {
                u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8E835',
                },
              u'4': {
                     u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8E8B1',
                     },
              u'6': {
                     u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8E7B3',
                     }},
            u'ipmi_address': u'10.49.27.28', u'bios_model': u'0824',
            u'default_vm_location': None, u'hypervisor_type': u'kKvm',
            u'service_vmexternal_ip': u'10.53.97.192',
            u'controller_vm_backplane_ip': u'10.53.97.192'},
 u'54830446-b55e-4f16-aa74-7b6a9ac9a7a4':
  {u'oplog_disk_pct': 3.4,
            u'memory_capacity_in_bytes': 135009402880, u'has_csr': False,
            u'default_vm_storage_container_uuid': None,
            u'hypervisor_username': u'root',
            u'service_vmnat_ip': None, u'hypervisor_key': u'10.53.97.187',
            u'acropolis_connection_state': u'kConnected',
            u'hypervisor_state': u'kAcropolisNormal',
            u'num_cpu_threads': 32, u'monitored': True,
            u'uuid': u'54830446-b55e-4f16-aa74-7b6a9ac9a7a4',
            u'reboot_pending': False, u'cpu_capacity_in_hz': 38384000000,
            u'num_cpu_sockets': 2, u'host_maintenance_mode_reason': None,
            u'hypervisor_address': u'10.53.97.187', u'host_gpus': None,
            u'failover_cluster_node_state': None, u'state': u'NORMAL',
            u'num_cpu_cores': 16, u'block_model': u'UseLayout',
            'guest_list':
             [{u'vm_features':
                {u'AGENT_VM': False, u'VGA_CONSOLE': True},
               u'name': u'PC', u'num_cores_per_vcpu': 1,
               u'gpus_assigned': False, u'num_vcpus': 4,
               u'memory_mb': 16384, u'power_state': u'on',
               u'ha_priority': 0, u'allow_live_migrate': True,
               u'timezone': u'UTC', u'vm_logical_timestamp': 10,
               u'host_uuid': u'54830446-b55e-4f16-aa74-7b6a9ac9a7a4',
               u'uuid': u'd90b5443-97f0-47eb-986d-f14e062448d4'},
              {u'vm_features':
                {u'AGENT_VM': False, u'VGA_CONSOLE': True},
               u'name': u'am1', u'num_cores_per_vcpu': 1,
               u'gpus_assigned': False, u'num_vcpus': 2,
               u'memory_mb': 4096, u'power_state': u'on',
               u'ha_priority': 0, u'allow_live_migrate': True,
               u'timezone': u'America/Los_Angeles',
               u'vm_logical_timestamp': 14,
               u'host_uuid': u'54830446-b55e-4f16-aa74-7b6a9ac9a7a4',
               u'uuid': u'0af0a010-0ad0-4fba-aa33-7cc3d0b6cb7e'}],
            u'cpu_model': u'Intel(R) Xeon(R) CPU E5-2630 v3 @ 2.40GHz',
            u'ipmi_username': u'ADMIN',
            u'service_vmid': u'0005809e-62e4-75c7-611b-0cc47ac3b354::6',
            u'bmc_model': u'X10_ATEN', u'host_nic_ids': [],
            u'cluster_uuid': u'0005809e-62e4-75c7-611b-0cc47ac3b354',
            u'stats':
             {u'num_read_io': u'27',
              u'controller_read_io_bandwidth_kBps': u'0',
              u'content_cache_hit_ppm': u'1000000',
             }, u'backplane_ip': None,
            u'vzone_name': u'', u'default_vhd_location': None,
            u'metadata_store_status_message': u'Metadata store enabled on the node',
            u'num_vms': 3, u'default_vm_storage_container_id': None,
            u'metadata_store_status': u'kNormalMode', u'name': u'foyt-3',
            u'hypervisor_password': None, u'service_vmnat_port': None,
            u'hypervisor_full_name': u'Nutanix 20180802.100874',
            u'is_degraded': False, u'host_type': u'HYPER_CONVERGED',
            u'default_vhd_storage_container_uuid': None,
            u'block_serial': u'15SM60250038',
            u'disk_hardware_configs':
             {u'1':
               {
                u'mount_path': u'/home/nutanix/data/stargate-storage/disks/BTHC506101ST480MGN',
                },
              u'3':
               {
                u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8DXYY',
                },
              u'2': {
                     u'mount_path': u'/home/nutanix/data/stargate-storage/disks/BTHC506101D4480MGN',
                     },
              u'5': {
                     u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8DQRM',
                     },
              u'4': {
                     u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8DJ7E',
                     },
              u'6': {
                     u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG8DVGG',
                     }},
            u'ipmi_address': u'10.49.27.27', u'bios_model': u'0824',
            u'hypervisor_type': u'kKvm',
            u'service_vmexternal_ip': u'10.53.97.191',
            u'controller_vm_backplane_ip': u'10.53.97.191'
 },
 u'acc819fe-e0ff-4963-93a4-5a0e1d3c77d3':
           {u'oplog_disk_pct': 3.4,
            u'memory_capacity_in_bytes': 270302969856, u'has_csr': False,
            u'default_vm_storage_container_uuid': None,
            u'hypervisor_username': u'root',
            u'key_management_device_to_certificate_status': {},
            u'service_vmnat_ip': None, u'hypervisor_key': u'10.53.96.75',
            u'acropolis_connection_state': u'kConnected',
            u'management_server_name': u'10.53.96.75',
            u'failover_cluster_fqdn': None, u'serial': u'ZM162S002621',
            u'bmc_version': u'01.97',
            u'hba_firmwares_list':
             [{u'hba_model': u'LSI Logic SAS3008',
               u'hba_version': u'MPTFW-10.00.03.00-IT'}],
            u'hypervisor_state': u'kAcropolisNormal',
            u'num_cpu_threads': 32, u'monitored': True,
            u'uuid': u'acc819fe-e0ff-4963-93a4-5a0e1d3c77d3',
            u'num_cpu_sockets': 2, u'host_maintenance_mode_reason': None,
            u'hypervisor_address': u'10.53.96.75',
            'guest_list':
             [{u'vm_features': {u'AGENT_VM': False, u'VGA_CONSOLE': True},
               u'name': u'am_RH_satellite', u'num_cores_per_vcpu': 2,
               u'gpus_assigned': False, u'num_vcpus': 4, u'memory_mb': 16384,
               u'power_state': u'on', u'ha_priority': 0,
               u'allow_live_migrate': True, u'timezone': u'America/Los_Angeles',
               u'vm_logical_timestamp': 5,
               u'host_uuid': u'acc819fe-e0ff-4963-93a4-5a0e1d3c77d3',
               u'uuid': u'e30f381d-d4bc-4958-a88c-79448efe5112'},
              {u'vm_features': {u'AGENT_VM': False, u'VGA_CONSOLE': True},
               u'name': u'am4', u'num_cores_per_vcpu': 1,
               u'gpus_assigned': False, u'num_vcpus': 2, u'memory_mb': 4096,
               u'power_state': u'on', u'ha_priority': 0,
               u'allow_live_migrate': True, u'timezone': u'America/Los_Angeles',
               u'vm_logical_timestamp': 2,
               u'host_uuid': u'acc819fe-e0ff-4963-93a4-5a0e1d3c77d3',
               u'uuid': u'f1e3362b-0377-4d70-bccd-63d2a1c09225'}],
            u'dynamic_ring_changing_node': None,
            u'cpu_model': u'Intel(R) Xeon(R) CPU E5-2630 v3 @ 2.40GHz',
            u'ipmi_username': u'ADMIN',
            u'cluster_uuid': u'0005809e-62e4-75c7-611b-0cc47ac3b354',
            u'ipmi_password': None, u'cpu_frequency_in_hz': 2400000000,
            u'stats': {u'num_read_io': u'47',
                       u'controller_read_io_bandwidth_kBps': u'0',
                       u'content_cache_hit_ppm': u'1000000',
                       }, u'backplane_ip': None,
            u'num_vms': 3,
            u'name': u'watermelon02-4', u'hypervisor_password': None,
            u'hypervisor_full_name': u'Nutanix 20180802.100874',
            u'is_degraded': False, u'host_type': u'HYPER_CONVERGED',
            u'default_vhd_storage_container_uuid': None,
            u'block_serial': u'16AP60170033', u'usage_stats': {
            },
            u'disk_hardware_configs': {
             u'1': {
                    u'mount_path': u'/home/nutanix/data/stargate-storage/disks/BTHC549209M3480MGN',
                    },
             u'3': {
                    u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG9TRZQ'},
             u'2': {
                    u'mount_path': u'/home/nutanix/data/stargate-storage/disks/BTHC550503XF480MGN',
                    },
             u'5': {
                    u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG9TS0N',
                    },
             u'4': {
                    u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG9TSF7',
                    },
             u'6': {
                    u'mount_path': u'/home/nutanix/data/stargate-storage/disks/9XG9TREW',
                    }}, u'ipmi_address': u'10.49.26.188',
            u'bios_model': u'0824', u'default_vm_location': None,
            u'hypervisor_type': u'kKvm',
            u'position': {u'ordinal': 4, u'physical_position': None,
                          u'name': u''},
            u'service_vmexternal_ip': u'10.53.96.79',
            u'controller_vm_backplane_ip': u'10.53.96.79'}}


class TestAhvConfigSection(TestBase):
    """
    Test base for testing class AhvConfigSection.
    """

    def __init__(self, *args, **kwargs):
        super(TestAhvConfigSection, self).__init__(*args, **kwargs)
        self.ahv_config = None

    def init_virt_config_section(self, is_pc=False):
        """
        Method executed before each unit test.
        """
        self.ahv_config = AhvConfigSection(MY_SECTION_NAME, None)
        if is_pc:
            self.ahv_config['prism_central'] = True
        # We need to set values using this way, because we need
        # to trigger __setitem__ of virt_config.
        for key, value in PE_SECTION_VALUES.items():
            self.ahv_config[key] = value

    def test_validate_ahv_PE_config(self):
        """
        Test validation of ahv section.
        """
        # PE validation.
        self.init_virt_config_section()
        result = self.ahv_config.validate()
        self.assertEqual(len(result), 0)

        # PC validation.
        self.init_virt_config_section(is_pc=True)
        result = self.ahv_config.validate()
        self.assertEqual(len(result), 0)

    def test_validate_ahv_invalid_server_ip(self):
        """
        Test validation of ahv config. Invalid server IP.
        """
        self.init_virt_config_section()
        self.ahv_config['server'] = '10.0.0.'
        result = self.ahv_config.validate()
        expected_result = ['Invalid server IP address provided']
        six.assertCountEqual(self, expected_result, result)

    def test_validate_ahv_config_missing_username_password(self):
        """
        Test validation of ahv config. Username and password is required.
        """
        self.init_virt_config_section()
        del self.ahv_config['username']
        del self.ahv_config['password']
        result = self.ahv_config.validate()
        expected_result = [
            ('error', 'Required option: "username" not set.'),
            ('error', 'Required option: "password" not set.')
        ]
        six.assertCountEqual(self, expected_result, result)

    def test_validate_ahv_config_invalid_internal_debug_flag(self):
        """
        Test validation of ahv config. If update_interval and internal debug
        are not set then we get a warning message for each flag.
        """
        self.init_virt_config_section()
        self.ahv_config['update_interval'] = 40
        result = self.ahv_config.validate()
        message = "Interval value can't be lower than {min} seconds. " \
                  "Default value of {min} " \
                  "seconds will be used.".format(min=DefaultUpdateInterval)
        expected_result = [("warning", message)]
        six.assertCountEqual(self, expected_result, result)


class TestAhv(TestBase):

    @staticmethod
    def create_config(name, wrapper, **kwargs):
        config = AhvConfigSection(name, wrapper)
        config.update(**kwargs)
        config.validate()
        return config

    def setUp(self, is_pc=False):
        config = self.create_config(name='test', wrapper=None, type='ahv',
                                    server='10.10.10.10', username='username',
                                    password='password', owner='owner',
                                    prism_central=is_pc)
        self.ahv = Virt.from_config(self.logger, config, Datastore(),
                                    interval=DefaultInterval)

    def run_once(self, queue=None):
        """Run AHV in oneshot mode."""
        self.ahv._oneshot = True
        self.ahv.dest = queue or Queue()
        self.ahv._terminate_event = Event()
        self.ahv._oneshot = True
        self.ahv._interval = 0
        self.ahv._run()

    @patch.object(Session, 'get')
    def test_connect_PE(self, mock_get):
        mock_get.return_value.status_code = 200
        self.run_once()

        self.assertEqual(mock_get.call_count, 3)
        call_list = [
            call('https://10.10.10.10:9440/api/nutanix/v2.0/clusters',
                 data=ANY, headers=ANY, timeout=ANY, verify=ANY),
            call('https://10.10.10.10:9440/api/nutanix/v2.0/vms',
                 data=ANY, headers=ANY, timeout=ANY, verify=ANY),
            call('https://10.10.10.10:9440/api/nutanix/v2.0/hosts',
                 data=ANY, headers=ANY, timeout=ANY, verify=ANY)
        ]
        mock_get.assert_has_calls(call_list, any_order=True)

    @patch.object(Session, 'post')
    def test_connect_PC(self, mock_post):
        self.setUp(is_pc=True)

        mock_post.return_value.status_code = 200
        self.run_once()

        self.assertEqual(mock_post.call_count, 3)
        call_list = [
            call('https://10.10.10.10:9440/api/nutanix/v3/clusters/list',
                 data=ANY, headers=ANY, timeout=ANY, verify=ANY),
            call('https://10.10.10.10:9440/api/nutanix/v3/vms/list',
                 data=ANY, headers=ANY, timeout=ANY, verify=ANY),
            call('https://10.10.10.10:9440/api/nutanix/v3/hosts/list',
                 data=ANY, headers=ANY, timeout=ANY, verify=ANY)
        ]
        mock_post.assert_has_calls(call_list, any_order=True)

    @patch.object(Session, 'get')
    def test_invalid_login_PE(self, mock_get):
        mock_get.return_value.ok = False
        mock_get.return_value.status_code = 401
        self.assertRaises(VirtError, self.run_once)

        mock_get.return_value.status_code = 403
        self.assertRaises(VirtError, self.run_once)

    @patch.object(Session, 'post')
    def test_invalid_login_PC(self, mock_post):
        self.setUp(is_pc=True)
        mock_post.return_value.ok = False
        mock_post.return_value.status_code = 401
        self.assertRaises(VirtError, self.run_once)

        mock_post.return_value.status_code = 403
        self.assertRaises(VirtError, self.run_once)

    @patch.object(Session, 'get')
    def test_connection_conflict_PE(self, mock_get):
        mock_get.return_value.ok = False
        mock_get.return_value.status_code = 409
        self.assertRaises(VirtError, self.run_once)

    @patch.object(Session, 'post')
    def test_connection_conflict_PC(self, mock_post):
        self.setUp(is_pc=True)
        mock_post.return_value.ok = False
        mock_post.return_value.status_code = 409
        self.assertRaises(VirtError, self.run_once)

    @patch.object(Session, 'get')
    def test_no_retry_http_erros_PE(self, mock_get):
        mock_get.return_value.ok = False
        mock_get.return_value.status_code = 400
        mock_get.return_value.text = 'Bad Request'
        self.assertRaises(VirtError, self.run_once)

        mock_get.return_value.status_code = 404
        mock_get.return_value.text = 'Not Found Error'
        self.assertRaises(VirtError, self.run_once)

        mock_get.return_value.status_code = 500
        mock_get.return_value.text = 'Internal Server Error'
        self.assertRaises(VirtError, self.run_once)

        mock_get.return_value.status_code = 502
        mock_get.return_value.tex = 'Bad Gateway'
        self.assertRaises(VirtError, self.run_once)

        mock_get.return_value.status_code = 503
        mock_get.return_value.text = 'Service Unavailable '
        self.assertRaises(VirtError, self.run_once)

    @patch.object(Session, 'post')
    def test_no_retry_http_erros_PC(self, mock_post):
        self.setUp(is_pc=True)
        mock_post.return_value.ok = False
        mock_post.return_value.status_code = 400
        mock_post.return_value.text = 'Bad Request'
        self.assertRaises(VirtError, self.run_once)

        mock_post.return_value.status_code = 404
        mock_post.return_value.text = 'Not Found Error'
        self.assertRaises(VirtError, self.run_once)

        mock_post.return_value.status_code = 500
        mock_post.return_value.text = 'Internal Server Error'
        self.assertRaises(VirtError, self.run_once)

        mock_post.return_value.status_code = 502
        mock_post.return_value.tex = 'Bad Gateway'
        self.assertRaises(VirtError, self.run_once)

        mock_post.return_value.status_code = 503
        mock_post.return_value.text = 'Service Unavailable '
        self.assertRaises(VirtError, self.run_once)

    @patch('virtwho.virt.ahv.ahv_interface.AhvInterface.build_host_to_uvm_map')
    def test_getHostGuestMapping(self, host_to_uvm_map):
        host_to_uvm_map.return_value = HOST_UVM_MAP

        expected_result = []

        for host_uuid in HOST_UVM_MAP:
           host = HOST_UVM_MAP[host_uuid]
           hypervisor_id = host_uuid
           host_name = host['name']
           cluster_uuid = host['cluster_uuid']
           guests = []
           for guest_vm in host['guest_list']:
               state = guest_vm['power_state']
               guests.append(Guest(guest_vm['uuid'], self.ahv.CONFIG_TYPE,
                                   state))

           facts = {
            Hypervisor.CPU_SOCKET_FACT: '2',
            Hypervisor.HYPERVISOR_TYPE_FACT: u'kKvm',
            Hypervisor.HYPERVISOR_VERSION_FACT: 'Nutanix 20180802.100874',
            Hypervisor.HYPERVISOR_CLUSTER: str(cluster_uuid)
           }

           expected_result.append(Hypervisor(
            name=host_name,
            hypervisorId=hypervisor_id,
            guestIds=guests,
            facts=facts
           ))

        result = self.ahv.getHostGuestMapping()['hypervisors']

        self.assertEqual(len(result), len(expected_result), 'lists length '
                                                            'do not match')
        for index in range(0, len(result)):
          self.assertEqual(expected_result[index].toDict(),
                           result[index].toDict())


import socket

from . import ahv_constants
from .ahv_interface import AhvInterface, Failure
from time import time
from virtwho import virt
from virtwho.config import VirtConfigSection
from virtwho.virt import Hypervisor, Guest

DefaultUpdateInterval = 1800
MinimumUpdateInterval = 60

class Ahv(virt.Virt):
  "AHV Rest client"
  CONFIG_TYPE = "ahv"
  def __init__(self, logger, config, dest, interval=None,
                terminate_event=None, oneshot=False):
    """
    Args:
      logger (Logger): Framework logger.
      config (onfigSection): Virtwho configuration.
      dest (Datastore): Data store for destination.
      interval (Int): Wait interval for continuous run.
      terminate_event (Event): Event on termination.
      one_shot (bool): Flag to run virtwho as onetime or continuously.
    Returns:
      None.
    """
    super(Ahv, self).__init__(logger, config, dest,
                                  terminate_event=terminate_event,
                                  interval=interval,
                                  oneshot=oneshot)
    self.config = config
    self.version = ahv_constants.VERSION_2
    self.is_pc = False
    if 'prism_central' in self.config:
      if self.config['prism_central']:
        self.version = ahv_constants.VERSION_3
        self.is_pc = True

    self.port = ahv_constants.DEFAULT_PORT
    self.url = ahv_constants.SERVER_BASE_URIL % (self.config['server'],
                                                 self.port, self.version)
    self.port = ahv_constants.DEFAULT_PORT
    self.username = self.config['username']
    self.password = self.config['password']
    self.update_interval = self.config['update_interval']
    self._interface = AhvInterface(logger, self.url, self.username,
                                   self.password, self.port,
                                   internal_debug=self.config['internal_debug'])

  def prepare(self):
    """
    Prepare for obtaining information from AHV server.
    Args:
      None
    Returns:
      None
    """
    self.logger.debug("Logging into Acropolis server %s" % self.url)
    self._interface.login(self.version)

  def _wait_for_update(self, timeout):
    """
    Wait for an update from AHV.
    Args:
      timeout (int): timeout
    Returns:
      task list (list): List of vm or host related tasks.
    """
    try:
      end_time = time() + timeout
      timestamp = int(time() * 1e6)
      while time() < end_time and not self.is_terminated():
        try:
          response = self._interface.get_tasks(timestamp, self.version,
                                               self.is_pc)
          if len(response) == 0:
            # No events, continue to wait
            continue
          self.logger.debug('AHV event found: %s\n' % response)
          return response
        except Failure as e:
          if 'timeout' not in e.details:
            raise
    except Exception:
      self.logger.exception("Waiting on AHV events failed: ")

    return []

  def getHostGuestMapping(self):
    """
    Get a dict of host to uvm mapping.
    Args:
      None.
    Returns:
      None.
    """
    mapping = {'hypervisors': []}
    
    host_uvm_map = self._interface.build_host_to_uvm_map(self.version)

    for host_uuid in host_uvm_map:
      host = host_uvm_map[host_uuid]
    
      try: 
        if self.config['hypervisor_id'] == 'uuid':
          hypervisor_id = host_uuid
        elif self.config['hypervisor_id'] == 'hostname':
          hypervisor_id = host['name']

      except KeyError:
        self.logger.debug("Host '%s' doesn't have hypervisor_id property",
                          host_uuid)
        continue

      guests = []
      if 'guest_list' in host and len(host['guest_list']) > 0:
        for guest_vm in host['guest_list']:
          try:
            state = guest_vm['power_state']
          except KeyError:
            self.logger.warning("Guest %s is missing power state. Perhaps they"
                                " are powered off", guest_vm['uuid'])
            continue
          guests.append(Guest(guest_vm['uuid'], self.CONFIG_TYPE, state))
      else:
        self.logger.debug("Host '%s' doesn't have any vms", host_uuid)

      cluster_uuid = self._interface.get_host_cluster_uuid(host)
      host_version = self._interface.get_host_version(host)
      host_name = host['name']

      facts = {
        Hypervisor.CPU_SOCKET_FACT: str(host['num_cpu_sockets']),
        Hypervisor.HYPERVISOR_TYPE_FACT: host.get('hypervisor_type', 'AHV'),
        Hypervisor.HYPERVISOR_VERSION_FACT: str(host_version),
        Hypervisor.HYPERVISOR_CLUSTER: str(cluster_uuid)}

      mapping['hypervisors'].append(virt.Hypervisor(hypervisorId=hypervisor_id,
                                                    guestIds=guests,
                                                    name=host_name,
                                                    facts=facts))
    return mapping

  def _run(self):
    """
    Continuous run loop for virt-who on AHV.
    Args:
      None.
    Returns:
      None.
    """
    self.prepare()
    next_update = time()
    initial = True
    wait_result = None
    while self._oneshot or not self.is_terminated():

      delta = next_update - time()

      if initial:
        assoc = self.getHostGuestMapping()
        self._send_data(virt.HostGuestAssociationReport(self.config, assoc))
        initial = False
        continue

      if delta > 0:
        # Wait for update.
        wait_result = self._wait_for_update(60 if initial else delta)
        if wait_result:
          events = wait_result
        else:
          events = []
      else:
        events = []

      if len(events) > 0 or delta > 0:
        assoc = self.getHostGuestMapping()
        self._send_data(virt.HostGuestAssociationReport(self.config, assoc))

      if self._oneshot:
        break
      else:
        next_update = time() + self.update_interval

class AhvConfigSection(VirtConfigSection):
  """Class for intializing and processing AHV config"""
  VIRT_TYPE = 'ahv'
  HYPERVISOR_ID = ('uuid', 'hwuuid', 'hostname')

  def __init__(self, *args, **kwargs):
    """
    Initialize AHV config and add config keys.
    Args:
       args: args
       kwargs : kwargs
     Returns:
       None.
    """
    super(AhvConfigSection, self).__init__(*args, **kwargs)
    self.add_key('server', validation_method=self._validate_server,
                 required=True)
    self.add_key('username', validation_method=self._validate_username,
                 required=True)
    self.add_key('password',
                 validation_method=self._validate_unencrypted_password,
                 required=True)
    self.add_key('is_hypervisor', validation_method=self._validate_str_to_bool,
                 default=True)
    self.add_key('prism_central', validation_method=self._validate_str_to_bool,
                 default=None)
    self.add_key('internal_debug', validation_method=self._validate_str_to_bool,
                 default=False)
    self.add_key('update_interval',
                 validation_method=self._validate_update_interval,
                 default=DefaultUpdateInterval)

  def _validate_server(self, key):
    """
    Validate the server IP address.
    Args:
      key (Str): server Ip address.
    Returns:
      Socket error is returned in case of an invalid ip.
    """
    error = super(AhvConfigSection, self)._validate_server(key)
    try:
      ip = self._values[key]
      socket.inet_aton(ip)
    except socket.error:
      error = 'Invalid server IP address provided'
    return error

  def _validate_update_interval(self, key):
    """
    Validate the update internal flag.
    Args:
      key (Int): Update internal value.
    Returns:
      A warning is returned in case interval is not valid.
    """
    result = None
    try:
      self._values[key] = int(self._values[key])

      if self._values[key] < MinimumUpdateInterval:
        message = "Interval value can't be lower than {min} seconds. " \
                  "Default value of {min} " \
                  "seconds will be used.".format(min=DefaultUpdateInterval)
        result = ("warning", message)
        self._values['interval'] = DefaultUpdateInterval
    except KeyError:
      result = ('warning', '%s is missing' % key)
    except (TypeError, ValueError) as e:
      result = (
      'warning', '%s was not set to a valid integer: %s' % (key, str(e)))
    return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

from .manager import Manager, ManagerError, ManagerFatalError, ManagerThrottleError

__all__ = ['Manager', 'ManagerError', 'ManagerFatalError', 'ManagerThrottleError']



from __future__ import print_function

import time

from fake_virt import FakeVirt, FakeHandler


class RhevmHandler(FakeHandler):
    def do_GET(self):
        time.sleep(0.1)
        print(("DO GET", self.path))
        if self.path == '/api':
            self.write_file('rhevm', 'rhev3_api.xml')
        elif self.path == '/api/clusters':
            self.write_file('rhevm', 'rhevm_clusters.xml')
        elif self.path == '/api/hosts':
            self.write_file('rhevm', 'rhevm_hosts.xml')
        elif self.path == '/api/vms':
            self.write_file('rhevm', 'rhevm_vms_%d.xml' % self.server._data_version.value)


class FakeRhevm(FakeVirt):
    def __init__(self, port=None):
        super(FakeRhevm, self).__init__(RhevmHandler, port=port)
        self.server._data_version = self._data_version

if __name__ == '__main__':
    rhevm = FakeRhevm(port=8443)
    rhevm.run()



# -*- coding: utf-8 -*-
from __future__ import print_function
"""
Module for communication with RHEV-M, part of virt-who

Copyright (C) 2012 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from six.moves import urllib
import requests
from requests.auth import HTTPBasicAuth
from xml.etree import ElementTree

from virtwho import virt
from virtwho.config import VirtConfigSection


RHEVM_STATE_TO_GUEST_STATE = {
    'unassigned': virt.Guest.STATE_UNKNOWN,
    'down': virt.Guest.STATE_SHUTOFF,
    'up': virt.Guest.STATE_RUNNING,
    'powering_up': virt.Guest.STATE_SHUTOFF,
    'powered_down': virt.Guest.STATE_SHUTINGDOWN,
    'paused': virt.Guest.STATE_PAUSED,
    'migrating_from': virt.Guest.STATE_SHUTOFF,
    'migrating_to': virt.Guest.STATE_SHUTOFF,
    'unknown': virt.Guest.STATE_UNKNOWN,
    'not_responding': virt.Guest.STATE_BLOCKED,
    'wait_for_launch': virt.Guest.STATE_BLOCKED,
    'reboot_in_progress': virt.Guest.STATE_SHUTOFF,
    'saving_state': virt.Guest.STATE_SHUTINGDOWN,
    'restoring_state': virt.Guest.STATE_SHUTOFF,
    'suspended': virt.Guest.STATE_PMSUSPENDED,
    'image_illegal': virt.Guest.STATE_CRASHED,
    'image_locked': virt.Guest.STATE_CRASHED,
    'powering_down': virt.Guest.STATE_SHUTINGDOWN
}


class RhevmConfigSection(VirtConfigSection):
    """
    This class is used for validation of RHEVM virtualization backend
    section. It tries to validate options and combination of options that
    are specific for this virtualization backend.
    """

    VIRT_TYPE = 'rhevm'
    HYPERVISOR_ID = ('uuid', 'hwuuid', 'hostname')

    def __init__(self, section_name, wrapper, *args, **kwargs):
        super(RhevmConfigSection, self).__init__(section_name, wrapper, *args, **kwargs)
        self.add_key('server', validation_method=self._validate_server, required=True)
        self.add_key('username', validation_method=self._validate_username, required=True)
        self.add_key('password', validation_method=self._validate_unencrypted_password, required=True)

    def _validate_server(self, key='server'):
        """
        Do validation of server option specific for this virtualization backend
        return: Return None or info/warning/error
        """
        url = self._values[key]

        if "//" not in url:
            url = "//" + url
        parsed = urllib.parse.urlsplit(url, "https")
        if ":" not in parsed[1]:
            netloc = parsed[1] + ":8443"
        else:
            netloc = parsed[1]
        url = urllib.parse.urlunsplit((parsed[0], netloc, parsed[2], "", ""))

        if url[-1] != '/':
            url += '/'

        if url != self._values[key]:
            self._values[key] = url
            return [(
                    'info',
                    "The original server URL was incomplete. It has been enhanced to %s" % url
                )]
        else:
            return None


class RhevM(virt.Virt):
    CONFIG_TYPE = "rhevm"

    def __init__(self, logger, config, dest, terminate_event=None,
                 interval=None, oneshot=False):
        super(RhevM, self).__init__(logger, config, dest,
                                    terminate_event=terminate_event,
                                    interval=interval,
                                    oneshot=oneshot)
        self.url = self.config['server']
        self.api_base = 'api'
        self.username = self.config['username']
        self.password = self.config['password']
        self.auth = HTTPBasicAuth(self.username.encode('utf-8'), self.password.encode('utf-8'))
        self.prepared = False
        self.clusters_url = None
        self.hosts_url = None
        self.vms_url = None

    def prepare(self):
        if not self.prepared:
            if not hasattr(self, 'major_version'):
                self.get_version()
            self.build_urls()
            self.prepared = True

    def build_urls(self):
        """
        Builds the URL's based on Rhev version
        """
        clusters_endpoint = '/clusters'
        hosts_endpoint = '/hosts'
        vms_endpoint = '/vms'

        self.clusters_url = urllib.parse.urljoin(self.url, self.api_base + clusters_endpoint)
        self.hosts_url = urllib.parse.urljoin(self.url, self.api_base + hosts_endpoint)
        self.vms_url = urllib.parse.urljoin(self.url, self.api_base + vms_endpoint)

    def get_version(self):
        """
        Gets the major version from the Rhevm server
        """
        try:
            headers = dict()
            headers['Version'] = '3'
            # We will store the api_base that seems to work and use that for future requests
            response = requests.get(urllib.parse.urljoin(self.url, self.api_base),
                                    auth=self.auth,
                                    headers=headers,
                                    verify=False)
            if response.status_code == 404:
                self.api_base = 'ovirt-engine/api'
                response = requests.get(urllib.parse.urljoin(self.url, self.api_base),
                                        auth=self.auth,
                                        headers=headers,
                                        verify=False)
            response.raise_for_status()
        except requests.RequestException as e:
            raise virt.VirtError("Unable to connect to RHEV-M server: %s" % str(e))

        try:
            api = ElementTree.fromstring(response.content)
        except Exception as e:
            self.logger.debug("Invalid xml file: %s" % response)
            raise virt.VirtError("Invalid XML file returned from RHEV-M: %s" % str(e))
        version = api.find('.//version')
        if version is not None:
            major = version.attrib['major']
            self.major_version = major
        else:
            self.logger.info("Could not determine version")

    def get(self, url):
        """
        Call RHEV-M server and retrieve what's on given url.
        """
        try:
            if self.major_version == '4':
                # If we are talking to a Rhev4 system, we need to specifically request
                # the Rhev 3 version of the api.  To minimize code impact, we do this
                # by setting a 'Version' header, as outlined in Rhev 4's "Version 3
                # REST API Guide"
                headers = dict()
                headers['Version'] = '3'
                response = requests.get(url, auth=self.auth, verify=False, headers=headers)
            else:
                response = requests.get(url, auth=self.auth, verify=False)
            response.raise_for_status()
        except requests.RequestException as e:
            raise virt.VirtError("Unable to connect to RHEV-M server: %s" % str(e))
        # FIXME: other errors
        return response.content

    def get_xml(self, url):
        """
        Call RHEV-M server, retrieve XML and parse it.
        """
        response = self.get(url)
        try:
            return ElementTree.fromstring(response)
        except Exception as e:
            self.logger.debug("Invalid xml file: %s" % response)
            raise virt.VirtError("Invalid XML file returned from RHEV-M: %s" % str(e))

    def getHostGuestMapping(self):
        """
        Returns dictionary containing a list of virt.Hypervisors
        Each virt.Hypervisor contains the hypervisor ID as well as a list of
        virt.Guest

        {'hypervisors': [Hypervisor1, ...]
        }
        """
        mapping = {}
        hosts = {}
        clusters = set()
        cluster_names = {}

        clusters_xml = self.get_xml(self.clusters_url)
        hosts_xml = self.get_xml(self.hosts_url)
        vms_xml = self.get_xml(self.vms_url)

        # Save ids of clusters that are "virt_service"
        for cluster in clusters_xml.findall('cluster'):
            cluster_id = cluster.get('id')
            virt_service = cluster.find('virt_service').text
            cluster_names[cluster_id] = cluster.find('name').text
            if virt_service.lower() == 'true':
                clusters.add(cluster_id)

        for host in hosts_xml.findall('host'):
            id = host.get('id')
            system_uuid = ''
            # Check if host is in cluster that is "virt_service"
            host_cluster = host.find('cluster')
            host_cluster_id = host_cluster.get('id')
            if host_cluster_id not in clusters:
                # Skip the host if it's cluster is not "virt_service"
                self.logger.debug('Cluster of host %s is not virt_service, skipped', id)
                continue

            try:
                system_uuid = host.find('hardware_information').find('uuid').text
            except AttributeError:
                # The error is not important yet
                self.logger.info("Unable to get hardware uuid for host %s ", id)


            if self.config['hypervisor_id'] == 'uuid':
                host_id = id
            elif self.config['hypervisor_id'] == 'hwuuid':
                if not system_uuid == '':
                    host_id = system_uuid
                else:
                    self.logger.error("Host %s doesn't have hardware uuid", id)
                    continue
            elif self.config['hypervisor_id'] == 'hostname':
                host_id = host.find('address').text

            sockets = host.find('cpu').find('topology').get('sockets')
            if not sockets:
                try:
                    sockets = host.find('cpu').find('topology').find('sockets').text
                except AttributeError:
                    sockets = "unknown"

            facts = {
                virt.Hypervisor.CPU_SOCKET_FACT: sockets,
                virt.Hypervisor.HYPERVISOR_TYPE_FACT: 'qemu',
                virt.Hypervisor.SYSTEM_UUID_FACT: system_uuid,
            }

            try:
                cluster_name = cluster_names[host_cluster_id]
            except KeyError:
                pass
            else:
                facts[virt.Hypervisor.HYPERVISOR_CLUSTER] = cluster_name

            try:
                version = host.find('version').get('full_version')
                if version:
                    facts[virt.Hypervisor.HYPERVISOR_VERSION_FACT] = version
            except AttributeError:
                pass

            hosts[id] = virt.Hypervisor(hypervisorId=host_id, name=host.find('address').text, facts=facts)
            mapping[id] = []
        for vm in vms_xml.findall('vm'):
            guest_id = vm.get('id')
            host = vm.find('host')
            if host is None:
                # Guest don't have any host
                continue

            host_id = host.get('id')
            if host_id not in mapping.keys():
                self.logger.warning(
                    "Guest %s claims that it belongs to host %s which doesn't exist",
                    guest_id, host_id)
                continue

            try:
                status = vm.find('status')
                try:
                    state_text = status.find('state').text.lower()
                except AttributeError:
                    # RHEVM 4.0 reports the state differently
                    state_text = status.text.lower()
                state = RHEVM_STATE_TO_GUEST_STATE.get(state_text, virt.Guest.STATE_UNKNOWN)
            except AttributeError:
                self.logger.warning(
                    "Guest %s doesn't report any status",
                    guest_id)
                state = virt.Guest.STATE_UNKNOWN

            hosts[host_id].guestIds.append(virt.Guest(guest_id, self.CONFIG_TYPE, state))

        return {'hypervisors': list(hosts.values())}

    def ping(self):
        return True

